"""
Manage software from FreeBSD ports

.. versionadded:: 2014.1.0

.. note::

    It may be helpful to use a higher timeout when running a
    :mod:`ports.installed <salt.states.ports>` state, since compiling the port
    may exceed Salt's timeout.

    .. code-block:: bash

        salt -t 1200 '*' state.highstate
"""

import copy
import logging

# Needed by imported function _options_file_exists
import os  # pylint: disable=W0611
import sys

import salt.utils.data
from salt.exceptions import CommandExecutionError, SaltInvocationError
from salt.modules.freebsdports import _normalize, _options_file_exists

log = logging.getLogger(__name__)


def __virtual__():
    if __grains__.get("os", "") == "FreeBSD" and "ports.install" in __salt__:
        return "ports"
    return (False, "ports module could not be loaded")


def _repack_options(options):
    """
    Repack the options data
    """
    return {
        str(x): _normalize(y)
        for x, y in salt.utils.data.repack_dictlist(options).items()
    }


def _get_option_list(options):
    """
    Returns the key/value pairs in the passed dict in a commaspace-delimited
    list in the format "key=value".
    """
    return ", ".join([f"{x}={y}" for x, y in options.items()])


def _build_option_string(options):
    """
    Common function to get a string to append to the end of the state comment
    """
    if options:
        return f"with the following build options: {_get_option_list(options)}"
    else:
        return "with the default build options"


def installed(name, options=None):
    """
    Verify that the desired port is installed, and that it was compiled with
    the desired options.

    options
        Make sure that the desired non-default options are set

        .. warning::

            Any build options not passed here assume the default values for the
            port, and are not just differences from the existing cached options
            from a previous ``make config``.

    Example usage:

    .. code-block:: yaml

        security/nmap:
          ports.installed:
            - options:
              - IPV6: off
    """
    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": f"{name} is already installed",
    }
    try:
        current_options = __salt__["ports.showconfig"](
            name, default=False, dict_return=True
        )
        default_options = __salt__["ports.showconfig"](
            name, default=True, dict_return=True
        )
        # unpack the options from the top-level return dict
        if current_options:
            current_options = current_options[next(iter(current_options))]
        if default_options:
            default_options = default_options[next(iter(default_options))]
    except (SaltInvocationError, CommandExecutionError) as exc:
        ret["result"] = False
        ret["comment"] = (
            "Unable to get configuration for {}. Port name may "
            "be invalid, or ports tree may need to be updated. "
            "Error message: {}".format(name, exc)
        )
        return ret

    options = _repack_options(options) if options is not None else {}
    desired_options = copy.deepcopy(default_options)
    desired_options.update(options)
    ports_pre = [
        x["origin"] for x in __salt__["pkg.list_pkgs"](with_origin=True).values()
    ]

    if current_options == desired_options and name in ports_pre:
        # Port is installed as desired
        if options:
            ret["comment"] += " " + _build_option_string(options)
        return ret

    if not default_options:
        if options:
            ret["result"] = False
            ret["comment"] = (
                "{} does not have any build options, yet options were specified".format(
                    name
                )
            )
            return ret
        else:
            if __opts__["test"]:
                ret["result"] = None
                ret["comment"] = f"{name} will be installed"
                return ret
    else:
        bad_opts = [x for x in options if x not in default_options]
        if bad_opts:
            ret["result"] = False
            ret["comment"] = (
                "The following options are not available for {}: {}".format(
                    name, ", ".join(bad_opts)
                )
            )
            return ret

        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"{name} will be installed "
            ret["comment"] += _build_option_string(options)
            return ret

        if options:
            if not __salt__["ports.config"](name, reset=True, **options):
                ret["result"] = False
                ret["comment"] = f"Unable to set options for {name}"
                return ret
        else:
            __salt__["ports.rmconfig"](name)
            if _options_file_exists(name):
                ret["result"] = False
                ret["comment"] = f"Unable to clear options for {name}"
                return ret

    ret["changes"] = __salt__["ports.install"](name)
    ports_post = [
        x["origin"] for x in __salt__["pkg.list_pkgs"](with_origin=True).values()
    ]
    err = sys.modules[__salt__["test.ping"].__module__].__context__.pop(
        "ports.install_error", None
    )
    if err or name not in ports_post:
        ret["result"] = False
    if ret["result"]:
        ret["comment"] = f"Successfully installed {name}"
        if default_options:
            ret["comment"] += " " + _build_option_string(options)
    else:
        ret["comment"] = f"Failed to install {name}"
        if err:
            ret["comment"] += f". Error message:\n{err}"
    return ret



"""
    tests.pytests.unit.beacons.test_swapusage
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Swap usage beacon test cases
"""

from collections import namedtuple

import pytest

import salt.beacons.swapusage as swapusage
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {}


@pytest.fixture
def stub_swap_usage():
    return namedtuple("sswap", "total used free percent sin sout")(
        17179865088,
        1674412032,
        15505453056,
        9.7,
        1572110336,
        3880046592,
    )


def test_non_list_config():
    config = {}

    ret = swapusage.validate(config)
    assert ret == (False, "Configuration for swapusage beacon must be a list.")


def test_empty_config():
    config = [{}]

    ret = swapusage.validate(config)
    assert ret == (False, "Configuration for swapusage beacon requires percent.")


def test_swapusage_match(stub_swap_usage):
    with patch("psutil.swap_memory", MagicMock(return_value=stub_swap_usage)):

        config = [{"percent": "9%"}, {"interval": 30}]

        ret = swapusage.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = swapusage.beacon(config)
        assert ret == [{"swapusage": 9.7}]

        # Test without the percent
        config = [{"percent": 9}, {"interval": 30}]

        ret = swapusage.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = swapusage.beacon(config)
        assert ret == [{"swapusage": 9.7}]


def test_swapusage_nomatch(stub_swap_usage):
    with patch("psutil.swap_memory", MagicMock(return_value=stub_swap_usage)):

        config = [{"percent": "10%"}]

        ret = swapusage.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = swapusage.beacon(config)
        assert ret != [{"swapusage": 9.7}]



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""

import pytest

import salt.states.nftables as nftables
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {nftables: {}}


def test_chain_present():
    """
    Test to verify the chain is exist.
    """
    ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}
    mock = MagicMock(
        side_effect=[
            {"result": True, "comment": ""},
            {"result": False, "comment": ""},
            {"result": False, "comment": ""},
        ]
    )
    with patch.dict(nftables.__salt__, {"nftables.check_chain": mock}):
        ret.update(
            {"comment": "nftables salt chain is already exist in filter table for ipv4"}
        )
        assert nftables.chain_present("salt") == ret

        mock = MagicMock(
            side_effect=[
                {"result": True, "comment": ""},
                {"result": False, "comment": ""},
            ]
        )
        with patch.dict(nftables.__salt__, {"nftables.new_chain": mock}):
            with patch.dict(nftables.__opts__, {"test": False}):
                ret.update(
                    {
                        "changes": {"locale": "salt"},
                        "comment": (
                            "nftables salt chain in filter table create success for"
                            " ipv4"
                        ),
                    }
                )
                assert nftables.chain_present("salt") == ret

                ret.update(
                    {
                        "changes": {},
                        "comment": (
                            "Failed to create salt chain in filter table:  for ipv4"
                        ),
                        "result": False,
                    }
                )
                assert nftables.chain_present("salt") == ret


def test_chain_absent():
    """
    Test to verify the chain is absent.
    """
    ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}
    mock = MagicMock(side_effect=[False, True])
    with patch.dict(nftables.__salt__, {"nftables.check_chain": mock}):
        ret.update(
            {
                "comment": (
                    "nftables salt chain is already absent in filter table for ipv4"
                )
            }
        )
        assert nftables.chain_absent("salt") == ret

        mock = MagicMock(return_value="")
        with patch.dict(nftables.__salt__, {"nftables.flush": mock}):
            ret.update(
                {
                    "result": False,
                    "comment": "Failed to flush salt chain in filter table:  for ipv4",
                }
            )
            assert nftables.chain_absent("salt") == ret


def test_append():
    """
    Test to append a rule to a chain
    """
    ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}
    mock = MagicMock(return_value=[])
    with patch.object(nftables, "_STATE_INTERNAL_KEYWORDS", mock):
        mock = MagicMock(return_value={"result": True, "comment": "", "rule": "a"})
        with patch.dict(nftables.__salt__, {"nftables.build_rule": mock}):
            mock = MagicMock(
                side_effect=[
                    {"result": True, "comment": ""},
                    {"result": False, "comment": ""},
                    {"result": False, "comment": ""},
                    {"result": False, "comment": ""},
                ]
            )
            with patch.dict(nftables.__salt__, {"nftables.check": mock}):
                ret.update(
                    {"comment": "nftables rule for salt already set (a) for ipv4"}
                )
                assert nftables.append("salt", table="", chain="") == ret

                with patch.dict(nftables.__opts__, {"test": True}):
                    ret.update(
                        {
                            "result": None,
                            "comment": (
                                "nftables rule for salt needs to be set (a) for ipv4"
                            ),
                        }
                    )
                    assert nftables.append("salt", table="", chain="") == ret

                with patch.dict(nftables.__opts__, {"test": False}):
                    mock = MagicMock(
                        side_effect=[
                            {"result": True, "comment": ""},
                            {"result": False, "comment": ""},
                        ]
                    )
                    with patch.dict(nftables.__salt__, {"nftables.append": mock}):
                        ret.update(
                            {
                                "changes": {"locale": "salt"},
                                "comment": "Set nftables rule for salt to: a for ipv4",
                                "result": True,
                            }
                        )
                        assert nftables.append("salt", table="", chain="") == ret

                        ret.update(
                            {
                                "changes": {},
                                "comment": (
                                    "Failed to set nftables rule for"
                                    " salt.\nAttempted rule was a for ipv4.\n"
                                ),
                                "result": False,
                            }
                        )
                        assert nftables.append("salt", table="", chain="") == ret


def test_insert():
    """
    Test to insert a rule into a chain
    """
    ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}
    mock = MagicMock(return_value=[])
    with patch.object(nftables, "_STATE_INTERNAL_KEYWORDS", mock):
        mock = MagicMock(return_value={"result": True, "comment": "", "rule": "a"})
        with patch.dict(nftables.__salt__, {"nftables.build_rule": mock}):
            mock = MagicMock(
                side_effect=[
                    {"result": True, "comment": ""},
                    {"result": False, "comment": ""},
                    {"result": False, "comment": ""},
                    {"result": False, "comment": ""},
                ]
            )
            with patch.dict(nftables.__salt__, {"nftables.check": mock}):
                ret.update(
                    {"comment": "nftables rule for salt already set for ipv4 (a)"}
                )
                assert nftables.insert("salt", table="", chain="") == ret

                with patch.dict(nftables.__opts__, {"test": True}):
                    ret.update(
                        {
                            "result": None,
                            "comment": (
                                "nftables rule for salt needs to be set for ipv4 (a)"
                            ),
                        }
                    )
                    assert nftables.insert("salt", table="", chain="") == ret

                with patch.dict(nftables.__opts__, {"test": False}):
                    mock = MagicMock(
                        side_effect=[
                            {"result": True, "comment": ""},
                            {"result": False, "comment": ""},
                        ]
                    )
                    with patch.dict(nftables.__salt__, {"nftables.insert": mock}):
                        ret.update(
                            {
                                "changes": {"locale": "salt"},
                                "comment": "Set nftables rule for salt to: a for ipv4",
                                "result": True,
                            }
                        )
                        assert (
                            nftables.insert("salt", table="", chain="", position="")
                            == ret
                        )

                        ret.update(
                            {
                                "changes": {},
                                "comment": (
                                    "Failed to set nftables rule for"
                                    " salt.\nAttempted rule was a"
                                ),
                                "result": False,
                            }
                        )
                        assert (
                            nftables.insert("salt", table="", chain="", position="")
                            == ret
                        )


def test_delete():
    """
    Test to delete a rule to a chain
    """
    ret = {"name": "salt", "changes": {}, "result": None, "comment": ""}

    mock = MagicMock(return_value=[])
    with patch.object(nftables, "_STATE_INTERNAL_KEYWORDS", mock):
        mock = MagicMock(return_value={"result": True, "comment": "", "rule": "a"})
        with patch.dict(nftables.__salt__, {"nftables.build_rule": mock}):
            mock = MagicMock(
                side_effect=[
                    {"result": False, "comment": ""},
                    {"result": True, "comment": ""},
                    {"result": True, "comment": ""},
                    {"result": True, "comment": ""},
                ]
            )
            with patch.dict(nftables.__salt__, {"nftables.check": mock}):
                ret.update(
                    {
                        "comment": "nftables rule for salt already absent for ipv4 (a)",
                        "result": True,
                    }
                )
                assert nftables.delete("salt", table="", chain="") == ret

                with patch.dict(nftables.__opts__, {"test": True}):
                    ret.update(
                        {
                            "result": None,
                            "comment": (
                                "nftables rule for salt needs to be deleted for"
                                " ipv4 (a)"
                            ),
                        }
                    )
                    assert nftables.delete("salt", table="", chain="") == ret

                with patch.dict(nftables.__opts__, {"test": False}):
                    mock = MagicMock(
                        side_effect=[
                            {"result": True, "comment": ""},
                            {"result": False, "comment": ""},
                        ]
                    )
                    with patch.dict(nftables.__salt__, {"nftables.delete": mock}):
                        ret.update(
                            {
                                "result": True,
                                "changes": {"locale": "salt"},
                                "comment": "Delete nftables rule for salt a",
                            }
                        )
                        assert (
                            nftables.delete("salt", table="", chain="", position="")
                            == ret
                        )

                        ret.update(
                            {
                                "result": False,
                                "changes": {},
                                "comment": (
                                    "Failed to delete nftables rule for"
                                    " salt.\nAttempted rule was a"
                                ),
                            }
                        )
                        assert (
                            nftables.delete("salt", table="", chain="", position="")
                            == ret
                        )


def test_flush():
    """
    Test to flush current nftables state
    """
    ret = {"name": "salt", "changes": {}, "result": None, "comment": ""}
    mock = MagicMock(return_value=[])
    with patch.object(nftables, "_STATE_INTERNAL_KEYWORDS", mock):
        mock = MagicMock(
            side_effect=[
                {"result": False, "comment": ""},
                {"result": True, "comment": ""},
                {"result": True, "comment": ""},
                {"result": True, "comment": ""},
            ]
        )
        with patch.dict(nftables.__salt__, {"nftables.check_table": mock}):
            with patch.dict(nftables.__opts__, {"test": False}):
                ret.update(
                    {
                        "comment": (
                            "Failed to flush table  in family ipv4, table does not"
                            " exist."
                        ),
                        "result": False,
                    }
                )
                assert (
                    nftables.flush("salt", table="", chain="", ignore_absence=False)
                    == ret
                )

                mock = MagicMock(
                    side_effect=[
                        {"result": False, "comment": ""},
                        {"result": True, "comment": ""},
                        {"result": True, "comment": ""},
                    ]
                )
                with patch.dict(nftables.__salt__, {"nftables.check_chain": mock}):
                    ret.update(
                        {
                            "comment": (
                                "Failed to flush chain  in table  in family ipv4, chain"
                                " does not exist."
                            )
                        }
                    )
                    assert (
                        nftables.flush("salt", table="", chain="", ignore_absence=False)
                        == ret
                    )

                    mock = MagicMock(
                        side_effect=[
                            {"result": True, "comment": ""},
                            {"result": False, "comment": ""},
                        ]
                    )
                    with patch.dict(nftables.__salt__, {"nftables.flush": mock}):
                        ret.update(
                            {
                                "changes": {"locale": "salt"},
                                "comment": (
                                    "Flush nftables rules in  table  chain ipv4 family"
                                ),
                                "result": True,
                            }
                        )
                        assert nftables.flush("salt", table="", chain="") == ret

                        ret.update(
                            {
                                "changes": {},
                                "comment": "Failed to flush nftables rules",
                                "result": False,
                            }
                        )
                        assert nftables.flush("salt", table="", chain="") == ret


def test_set_policy():
    """
    Test to sets the default policy for nftables firewall tables
    """
    ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}

    mock = MagicMock(return_value=[])
    with patch.object(nftables, "_STATE_INTERNAL_KEYWORDS", mock):
        mock = MagicMock(return_value="stack")
        with patch.dict(nftables.__salt__, {"nftables.get_policy": mock}):
            ret.update(
                {
                    "comment": (
                        "nftables default policy for chain"
                        "  on table  for ipv4 already set to stack"
                    )
                }
            )
            assert (
                nftables.set_policy("salt", table="", chain="", policy="stack") == ret
            )

            with patch.dict(nftables.__opts__, {"test": True}):
                ret.update(
                    {
                        "comment": (
                            "nftables default policy for chain  on table  for "
                            "ipv4 needs to be set to sal"
                        ),
                        "result": None,
                    }
                )
                assert (
                    nftables.set_policy("salt", table="", chain="", policy="sal") == ret
                )

            with patch.dict(nftables.__opts__, {"test": False}):
                mock = MagicMock(side_effect=[True, False])
                with patch.dict(nftables.__salt__, {"nftables.set_policy": mock}):
                    ret.update(
                        {
                            "changes": {"locale": "salt"},
                            "comment": "Set default policy for  to sal family ipv4",
                            "result": True,
                        }
                    )
                    assert (
                        nftables.set_policy("salt", table="", chain="", policy="sal")
                        == ret
                    )

                    ret.update(
                        {
                            "comment": "Failed to set nftables default policy",
                            "result": False,
                            "changes": {},
                        }
                    )
                    assert (
                        nftables.set_policy("salt", table="", chain="", policy="sal")
                        == ret
                    )


def test_table_present():
    """
    Test to verify a table exists.
    """
    ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}

    mock = MagicMock(
        side_effect=[
            {"result": True},
            {"result": False},
            {"result": False},
            {"result": False},
        ]
    )
    with patch.dict(nftables.__salt__, {"nftables.check_table": mock}):
        ret.update({"comment": "nftables table salt already exists in family ipv4"})
        assert nftables.table_present("salt") == ret

        with patch.dict(nftables.__opts__, {"test": True}):
            ret.update(
                {
                    "comment": "nftables table salt would be created in family ipv4",
                    "result": None,
                }
            )
            assert nftables.table_present("salt") == ret

        with patch.dict(nftables.__opts__, {"test": False}):
            mock = MagicMock(side_effect=[{"result": True}, {"result": False}])
            with patch.dict(nftables.__salt__, {"nftables.new_table": mock}):
                ret.update(
                    {
                        "result": True,
                        "comment": (
                            "nftables table salt successfully created in family ipv4"
                        ),
                        "changes": {"locale": "salt"},
                    }
                )
                assert nftables.table_present("salt") == ret

                ret.update(
                    {
                        "changes": {},
                        "result": False,
                        "comment": "Failed to create table salt for family ipv4",
                    }
                )
                assert nftables.table_present("salt") == ret


def test_table_absent():
    """
    Test to verify a table is absent.
    """
    ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}

    mock = MagicMock(
        side_effect=[
            {"result": False},
            {"result": True},
            {"result": True},
            {"result": True},
        ]
    )
    with patch.dict(nftables.__salt__, {"nftables.check_table": mock}):
        ret.update(
            {"comment": "nftables table salt is already absent from family ipv4"}
        )
        assert nftables.table_absent("salt") == ret

        with patch.dict(nftables.__opts__, {"test": True}):
            ret.update(
                {
                    "comment": "nftables table salt would be deleted from family ipv4",
                    "result": None,
                }
            )
            assert nftables.table_absent("salt") == ret

        with patch.dict(nftables.__opts__, {"test": False}):
            mock = MagicMock(side_effect=[False, "a"])
            with patch.dict(nftables.__salt__, {"nftables.flush": mock}):
                mock = MagicMock(side_effect=[{"result": True}, {"result": False}])
                with patch.dict(nftables.__salt__, {"nftables.delete_table": mock}):
                    ret.update(
                        {
                            "changes": {"locale": "salt"},
                            "comment": (
                                "nftables table salt successfully deleted from family"
                                " ipv4"
                            ),
                            "result": True,
                        }
                    )
                    assert nftables.table_absent("salt") == ret

                    ret.update(
                        {
                            "changes": {},
                            "result": False,
                            "comment": "Failed to delete table salt from family ipv4",
                        }
                    )
                    assert nftables.table_absent("salt") == ret



"""
Pillar include tests
"""

import pytest


@pytest.fixture(scope="module")
def pillar_include_tree(base_env_pillar_tree_root_dir, salt_minion, salt_call_cli):
    top_file = """
    base:
      '{}':
        - include
        - glob-include
        - include-c
        - include-d
    """.format(
        salt_minion.id
    )
    include_pillar_file = """
    include:
      - include-a:
          key: element:a
      - include-b:
          key: element:b
    """
    include_a_pillar_file = """
    a:
      - 'Entry A'
    """
    include_b_pillar_file = """
    b:
      - 'Entry B'
    """
    include_c_pillar_file = """
    c:
      - 'Entry C'
    """
    include_d_pillar_file = """
    include:
      - include-c:
          key: element:d
    """
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_pillar_tree_root_dir
    )
    include_tempfile = pytest.helpers.temp_file(
        "include.sls", include_pillar_file, base_env_pillar_tree_root_dir
    )
    include_a_tempfile = pytest.helpers.temp_file(
        "include-a.sls", include_a_pillar_file, base_env_pillar_tree_root_dir
    )
    include_b_tempfile = pytest.helpers.temp_file(
        "include-b.sls", include_b_pillar_file, base_env_pillar_tree_root_dir
    )
    include_c_tempfile = pytest.helpers.temp_file(
        "include-c.sls", include_c_pillar_file, base_env_pillar_tree_root_dir
    )
    include_d_tempfile = pytest.helpers.temp_file(
        "include-d.sls", include_d_pillar_file, base_env_pillar_tree_root_dir
    )
    glob_include_pillar_file = """
    include:
      - 'glob-include-*'
    """
    glob_include_a_pillar_file = """
    glob-a:
      - 'Entry A'
    """
    glob_include_b_pillar_file = """
    glob-b:
      - 'Entry B'
    """
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_pillar_tree_root_dir
    )
    glob_include_tempfile = pytest.helpers.temp_file(
        "glob-include.sls", glob_include_pillar_file, base_env_pillar_tree_root_dir
    )
    glob_include_a_tempfile = pytest.helpers.temp_file(
        "glob-include-a.sls", glob_include_a_pillar_file, base_env_pillar_tree_root_dir
    )
    glob_include_b_tempfile = pytest.helpers.temp_file(
        "glob-include-b.sls", glob_include_b_pillar_file, base_env_pillar_tree_root_dir
    )
    try:
        with (
            top_tempfile
        ), (
            include_tempfile
        ), (
            include_a_tempfile
        ), include_b_tempfile, include_c_tempfile, include_d_tempfile:
            with (
                glob_include_tempfile
            ), glob_include_a_tempfile, glob_include_b_tempfile:
                ret = salt_call_cli.run("saltutil.refresh_pillar", wait=True)
                assert ret.returncode == 0
                assert ret.data is True
                yield
    finally:
        # Refresh pillar again to cleaup the temp pillar
        ret = salt_call_cli.run("saltutil.refresh_pillar", wait=True)
        assert ret.returncode == 0
        assert ret.data is True


def test_pillar_include(pillar_include_tree, salt_call_cli):
    """
    Test pillar include
    """
    ret = salt_call_cli.run("pillar.items")
    assert ret.returncode == 0
    assert ret.data
    assert "element" in ret.data
    assert "a" in ret.data["element"]
    assert ret.data["element"]["a"] == {"a": ["Entry A"]}
    assert "b" in ret.data["element"]
    assert ret.data["element"]["b"] == {"b": ["Entry B"]}


def test_pillar_glob_include(pillar_include_tree, salt_call_cli):
    """
    Test pillar include via glob pattern
    """
    ret = salt_call_cli.run("pillar.items")
    assert ret.returncode == 0
    assert ret.data
    assert "glob-a" in ret.data
    assert ret.data["glob-a"] == ["Entry A"]
    assert "glob-b" in ret.data
    assert ret.data["glob-b"] == ["Entry B"]


def test_pillar_include_already_included(pillar_include_tree, salt_call_cli):
    """
    Test pillar include when a pillar file
    has already been included.
    """
    ret = salt_call_cli.run("pillar.items")
    assert ret.returncode == 0
    assert ret.data
    assert "element" in ret.data
    assert "d" in ret.data["element"]
    assert ret.data["element"]["d"] == {"c": ["Entry C"]}



import os

import pytest

import salt.modules.pkgin as pkgin
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules(tmp_path):
    return {pkgin: {"__opts__": {"cachedir": str(tmp_path)}}}


def test_search():
    """
    Test searching for a package
    """

    pkgin_out = [
        "somepkg-1.0          Some package description here",
        "",
        "=: package is installed and up-to-date",
        "<: package is installed but newer version is available",
        ">: installed package has a greater version than available package",
    ]

    pkgin__get_version_mock = MagicMock(return_value=["0", "9", "0"])
    pkgin__check_pkgin_mock = MagicMock(return_value="/opt/pkg/bin/pkgin")
    pkgin_search_cmd = MagicMock(return_value=os.linesep.join(pkgin_out))

    with patch("salt.modules.pkgin._get_version", pkgin__get_version_mock), patch(
        "salt.modules.pkgin._check_pkgin", pkgin__check_pkgin_mock
    ), patch.dict(pkgin.__salt__, {"cmd.run": pkgin_search_cmd}):
        assert pkgin.search("somepkg") == {"somepkg": "1.0"}

    pkgin_out = [
        "somepkg-1.0 =          Some package description here",
        "",
        "=: package is installed and up-to-date",
        "<: package is installed but newer version is available",
        ">: installed package has a greater version than available package",
    ]

    pkgin_search_cmd = MagicMock(return_value=os.linesep.join(pkgin_out))

    with patch("salt.modules.pkgin._get_version", pkgin__get_version_mock), patch(
        "salt.modules.pkgin._check_pkgin", pkgin__check_pkgin_mock
    ), patch.dict(pkgin.__salt__, {"cmd.run": pkgin_search_cmd}):
        assert pkgin.search("somepkg") == {"somepkg": "1.0"}


def test_latest_version():
    """
    Test getting the latest version of a package
    """

    pkgin_out = [
        "somepkg-1.0;;Some package description here",
        "",
        "=: package is installed and up-to-date",
        "<: package is installed but newer version is available",
        ">: installed package has a greater version than available package",
    ]

    pkgin__get_version_mock = MagicMock(return_value=["0", "9", "0"])
    pkgin__check_pkgin_mock = MagicMock(return_value="/opt/pkg/bin/pkgin")
    pkgin_refresh_db_mock = MagicMock(return_value=True)
    pkgin_search_cmd = MagicMock(return_value=os.linesep.join(pkgin_out))

    with patch("salt.modules.pkgin.refresh_db", pkgin_refresh_db_mock), patch(
        "salt.modules.pkgin._get_version", pkgin__get_version_mock
    ), patch("salt.modules.pkgin._check_pkgin", pkgin__check_pkgin_mock), patch.dict(
        pkgin.__salt__, {"cmd.run": pkgin_search_cmd}
    ):
        assert pkgin.latest_version("somepkg") == "1.0"

    pkgin_out = [
        "somepkg-1.1;<;Some package description here",
        "",
        "=: package is installed and up-to-date",
        "<: package is installed but newer version is available",
        ">: installed package has a greater version than available package",
    ]

    pkgin_refresh_db_mock = MagicMock(return_value=True)
    pkgin_search_cmd = MagicMock(return_value=os.linesep.join(pkgin_out))

    with patch("salt.modules.pkgin.refresh_db", pkgin_refresh_db_mock), patch(
        "salt.modules.pkgin._get_version", pkgin__get_version_mock
    ), patch("salt.modules.pkgin._check_pkgin", pkgin__check_pkgin_mock), patch.dict(
        pkgin.__salt__, {"cmd.run": pkgin_search_cmd}
    ):
        assert pkgin.latest_version("somepkg") == "1.1"

    pkgin_out = [
        "somepkg-1.2;=;Some package description here",
        "",
        "=: package is installed and up-to-date",
        "<: package is installed but newer version is available",
        ">: installed package has a greater version than available package",
    ]

    pkgin_refresh_db_mock = MagicMock(return_value=True)
    pkgin_search_cmd = MagicMock(return_value=os.linesep.join(pkgin_out))

    with patch("salt.modules.pkgin.refresh_db", pkgin_refresh_db_mock), patch(
        "salt.modules.pkgin._get_version", pkgin__get_version_mock
    ), patch("salt.modules.pkgin._check_pkgin", pkgin__check_pkgin_mock), patch.dict(
        pkgin.__salt__, {"cmd.run": pkgin_search_cmd}
    ):
        assert pkgin.latest_version("somepkg") == "1.2"

    pkgin_out = "No results found for ^boguspkg$"

    pkgin_refresh_db_mock = MagicMock(return_value=True)
    pkgin_search_cmd = MagicMock(return_value=pkgin_out)

    with patch("salt.modules.pkgin.refresh_db", pkgin_refresh_db_mock), patch(
        "salt.modules.pkgin._get_version", pkgin__get_version_mock
    ), patch("salt.modules.pkgin._check_pkgin", pkgin__check_pkgin_mock), patch.dict(
        pkgin.__salt__, {"cmd.run": pkgin_search_cmd}
    ):
        assert pkgin.latest_version("boguspkg") == {}


def test_file_dict():
    """
    Test that file_dict doesn't crash
    """
    pkg_info_stdout = [
        "/opt/pkg/bin/pkgin",
        "/opt/pkg/man/man1/pkgin.1",
        "/opt/pkg/share/examples/pkgin/preferred.conf.example",
        "/opt/pkg/share/examples/pkgin/repositories.conf.example",
    ]

    pkg_info_out = {
        "pid": 1234,
        "retcode": 0,
        "stderr": "",
        "stdout": os.linesep.join(pkg_info_stdout),
    }

    pkg_info_cmd = MagicMock(return_value=pkg_info_out)

    with patch.dict(pkgin.__salt__, {"cmd.run_all": pkg_info_cmd}):
        assert pkgin.file_dict("pkgin") == {
            "files": {
                "pkgin": [
                    "/opt/pkg/bin/pkgin",
                    "/opt/pkg/man/man1/pkgin.1",
                    "/opt/pkg/share/examples/pkgin/preferred.conf.example",
                    "/opt/pkg/share/examples/pkgin/repositories.conf.example",
                ]
            }
        }



import os.path
import random
import string
import sys

import pytest

import salt.config
import salt.states.boto_vpc as boto_vpc
import salt.utils.botomod as botomod
from salt.utils.versions import Version
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase

# pylint: disable=import-error,unused-import
from tests.unit.modules.test_boto_vpc import BotoVpcTestCaseMixin

try:
    import boto

    boto.ENDPOINTS_PATH = os.path.join(
        RUNTIME_VARS.TESTS_DIR, "unit/files/endpoints.json"
    )
    import boto3
    from boto.exception import BotoServerError

    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False

try:
    from moto import mock_ec2_deprecated  # pylint: disable=no-name-in-module

    HAS_MOTO = True
except ImportError:
    HAS_MOTO = False

    def mock_ec2_deprecated(self):
        """
        if the mock_ec2_deprecated function is not available due to import failure
        this replaces the decorated function with stub_function.
        Allows boto_vpc unit tests to use the @mock_ec2_deprecated decorator
        without a "NameError: name 'mock_ec2_deprecated' is not defined" error.
        """

        def stub_function(self):
            pass

        return stub_function


# pylint: enable=import-error,unused-import

# the boto_vpc module relies on the connect_to_region() method
# which was added in boto 2.8.0
# https://github.com/boto/boto/commit/33ac26b416fbb48a60602542b4ce15dcc7029f12
required_boto_version = "2.8.0"
region = "us-east-1"
access_key = "GKTADJGHEIQSXMKKRBJ08H"
secret_key = "askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs"
conn_parameters = {
    "region": region,
    "key": access_key,
    "keyid": secret_key,
    "profile": {},
}
cidr_block = "10.0.0.0/24"
subnet_id = "subnet-123456"
dhcp_options_parameters = {
    "domain_name": "example.com",
    "domain_name_servers": ["1.2.3.4"],
    "ntp_servers": ["5.6.7.8"],
    "netbios_name_servers": ["10.0.0.1"],
    "netbios_node_type": 2,
}
network_acl_entry_parameters = ("fake", 100, -1, "allow", cidr_block)
dhcp_options_parameters.update(conn_parameters)


def _has_required_boto():
    """
    Returns True/False boolean depending on if Boto is installed and correct
    version.
    """
    if not HAS_BOTO:
        return False
    elif Version(boto.__version__) < Version(required_boto_version):
        return False
    else:
        return True


class BotoVpcStateTestCaseBase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        ctx = {}
        utils = salt.loader.utils(
            self.opts,
            whitelist=["boto", "boto3", "args", "systemd", "path", "platform", "reg"],
            context=ctx,
        )
        serializers = salt.loader.serializers(self.opts)
        self.funcs = salt.loader.minion_mods(
            self.opts, context=ctx, utils=utils, whitelist=["boto_vpc", "config"]
        )
        self.salt_states = salt.loader.states(
            opts=self.opts,
            functions=self.funcs,
            utils=utils,
            whitelist=["boto_vpc"],
            serializers=serializers,
        )
        return {
            boto_vpc: {
                "__opts__": self.opts,
                "__salt__": self.funcs,
                "__utils__": utils,
                "__states__": self.salt_states,
                "__serializers__": serializers,
            },
            botomod: {},
        }

    @classmethod
    def setUpClass(cls):
        cls.opts = salt.config.DEFAULT_MINION_OPTS.copy()
        cls.opts["grains"] = salt.loader.grains(cls.opts)

    @classmethod
    def tearDownClass(cls):
        del cls.opts

    def setUp(self):
        self.addCleanup(delattr, self, "funcs")
        self.addCleanup(delattr, self, "salt_states")
        # connections keep getting cached from prior tests, can't find the
        # correct context object to clear it. So randomize the cache key, to prevent any
        # cache hits
        conn_parameters["key"] = "".join(
            random.choice(string.ascii_lowercase + string.digits) for _ in range(50)
        )


@pytest.mark.skipif(HAS_BOTO is False, reason="The boto module must be installed.")
@pytest.mark.skipif(HAS_MOTO is False, reason="The moto module must be installed.")
@pytest.mark.skipif(
    _has_required_boto() is False,
    reason="The boto module must be greater than or equal to version {}".format(
        required_boto_version
    ),
)
class BotoVpcTestCase(BotoVpcStateTestCaseBase, BotoVpcTestCaseMixin):
    """
    TestCase for salt.states.boto_vpc state.module
    """

    @pytest.mark.skipif(
        sys.version_info > (3, 6),
        reason="Disabled for 3.7+ pending https://github.com/spulec/moto/issues/1706.",
    )
    @mock_ec2_deprecated
    @pytest.mark.slow_test
    def test_present_when_vpc_does_not_exist(self):
        """
        Tests present on a VPC that does not exist.
        """
        with patch.dict(botomod.__salt__, self.funcs):
            vpc_present_result = self.salt_states["boto_vpc.present"](
                "test", cidr_block
            )

        self.assertTrue(vpc_present_result["result"])
        self.assertEqual(
            vpc_present_result["changes"]["new"]["vpc"]["state"], "available"
        )

    @pytest.mark.skipif(
        sys.version_info > (3, 6),
        reason="Disabled for 3.7+ pending https://github.com/spulec/moto/issues/1706.",
    )
    @mock_ec2_deprecated
    def test_present_when_vpc_exists(self):
        vpc = self._create_vpc(name="test")
        vpc_present_result = self.salt_states["boto_vpc.present"]("test", cidr_block)
        self.assertTrue(vpc_present_result["result"])
        self.assertEqual(vpc_present_result["changes"], {})

    @mock_ec2_deprecated
    @pytest.mark.skip(
        reason="Disabled pending https://github.com/spulec/moto/issues/493"
    )
    def test_present_with_failure(self):
        with patch(
            "moto.ec2.models.VPCBackend.create_vpc",
            side_effect=BotoServerError(400, "Mocked error"),
        ):
            vpc_present_result = self.salt_states["boto_vpc.present"](
                "test", cidr_block
            )
            self.assertFalse(vpc_present_result["result"])
            self.assertTrue("Mocked error" in vpc_present_result["comment"])

    @pytest.mark.skipif(
        sys.version_info > (3, 6),
        reason="Disabled for 3.7+ pending https://github.com/spulec/moto/issues/1706.",
    )
    @mock_ec2_deprecated
    @pytest.mark.slow_test
    def test_absent_when_vpc_does_not_exist(self):
        """
        Tests absent on a VPC that does not exist.
        """
        with patch.dict(botomod.__salt__, self.funcs):
            vpc_absent_result = self.salt_states["boto_vpc.absent"]("test")
        self.assertTrue(vpc_absent_result["result"])
        self.assertEqual(vpc_absent_result["changes"], {})

    @pytest.mark.skipif(
        sys.version_info > (3, 6),
        reason="Disabled for 3.7+ pending https://github.com/spulec/moto/issues/1706.",
    )
    @mock_ec2_deprecated
    @pytest.mark.slow_test
    def test_absent_when_vpc_exists(self):
        vpc = self._create_vpc(name="test")
        with patch.dict(botomod.__salt__, self.funcs):
            vpc_absent_result = self.salt_states["boto_vpc.absent"]("test")
        self.assertTrue(vpc_absent_result["result"])
        self.assertEqual(vpc_absent_result["changes"]["new"]["vpc"], None)

    @mock_ec2_deprecated
    @pytest.mark.skip(
        reason="Disabled pending https://github.com/spulec/moto/issues/493"
    )
    def test_absent_with_failure(self):
        vpc = self._create_vpc(name="test")
        with patch(
            "moto.ec2.models.VPCBackend.delete_vpc",
            side_effect=BotoServerError(400, "Mocked error"),
        ):
            vpc_absent_result = self.salt_states["boto_vpc.absent"]("test")
            self.assertFalse(vpc_absent_result["result"])
            self.assertTrue("Mocked error" in vpc_absent_result["comment"])


class BotoVpcResourceTestCaseMixin(BotoVpcTestCaseMixin):
    resource_type = None
    backend_create = None
    backend_delete = None
    extra_kwargs = {}

    def _create_resource(self, vpc_id=None, name=None):
        _create = getattr(self, "_create_" + self.resource_type)
        _create(vpc_id=vpc_id, name=name, **self.extra_kwargs)

    @pytest.mark.skipif(
        sys.version_info > (3, 6),
        reason="Disabled for 3.7+ pending https://github.com/spulec/moto/issues/1706.",
    )
    @mock_ec2_deprecated
    @pytest.mark.slow_test
    def test_present_when_resource_does_not_exist(self):
        """
        Tests present on a resource that does not exist.
        """
        vpc = self._create_vpc(name="test")
        with patch.dict(botomod.__salt__, self.funcs):
            resource_present_result = self.salt_states[
                f"boto_vpc.{self.resource_type}_present"
            ](name="test", vpc_name="test", **self.extra_kwargs)

        self.assertTrue(resource_present_result["result"])

        exists = self.funcs["boto_vpc.resource_exists"](self.resource_type, "test").get(
            "exists"
        )
        self.assertTrue(exists)

    @pytest.mark.skipif(
        sys.version_info > (3, 6),
        reason="Disabled for 3.7+ pending https://github.com/spulec/moto/issues/1706.",
    )
    @mock_ec2_deprecated
    @pytest.mark.slow_test
    def test_present_when_resource_exists(self):
        vpc = self._create_vpc(name="test")
        self._create_resource(vpc_id=vpc.id, name="test")
        with patch.dict(botomod.__salt__, self.funcs):
            resource_present_result = self.salt_states[
                f"boto_vpc.{self.resource_type}_present"
            ](name="test", vpc_name="test", **self.extra_kwargs)
        self.assertTrue(resource_present_result["result"])
        self.assertEqual(resource_present_result["changes"], {})

    @mock_ec2_deprecated
    @pytest.mark.skip(
        reason="Disabled pending https://github.com/spulec/moto/issues/493"
    )
    def test_present_with_failure(self):
        vpc = self._create_vpc(name="test")
        with patch(
            f"moto.ec2.models.{self.backend_create}",
            side_effect=BotoServerError(400, "Mocked error"),
        ):
            resource_present_result = self.salt_states[
                f"boto_vpc.{self.resource_type}_present"
            ](name="test", vpc_name="test", **self.extra_kwargs)

            self.assertFalse(resource_present_result["result"])
            self.assertTrue("Mocked error" in resource_present_result["comment"])

    @pytest.mark.skipif(
        sys.version_info > (3, 6),
        reason="Disabled for 3.7+ pending https://github.com/spulec/moto/issues/1706.",
    )
    @mock_ec2_deprecated
    @pytest.mark.slow_test
    def test_absent_when_resource_does_not_exist(self):
        """
        Tests absent on a resource that does not exist.
        """
        with patch.dict(botomod.__salt__, self.funcs):
            resource_absent_result = self.salt_states[
                f"boto_vpc.{self.resource_type}_absent"
            ]("test")
        self.assertTrue(resource_absent_result["result"])
        self.assertEqual(resource_absent_result["changes"], {})

    @pytest.mark.skipif(
        sys.version_info > (3, 6),
        reason="Disabled for 3.7+ pending https://github.com/spulec/moto/issues/1706.",
    )
    @mock_ec2_deprecated
    @pytest.mark.slow_test
    def test_absent_when_resource_exists(self):
        vpc = self._create_vpc(name="test")
        self._create_resource(vpc_id=vpc.id, name="test")

        with patch.dict(botomod.__salt__, self.funcs):
            resource_absent_result = self.salt_states[
                f"boto_vpc.{self.resource_type}_absent"
            ]("test")
        self.assertTrue(resource_absent_result["result"])
        self.assertEqual(
            resource_absent_result["changes"]["new"][self.resource_type], None
        )
        exists = self.funcs["boto_vpc.resource_exists"](self.resource_type, "test").get(
            "exists"
        )
        self.assertFalse(exists)

    @mock_ec2_deprecated
    @pytest.mark.skip(
        reason="Disabled pending https://github.com/spulec/moto/issues/493"
    )
    def test_absent_with_failure(self):
        vpc = self._create_vpc(name="test")
        self._create_resource(vpc_id=vpc.id, name="test")

        with patch(
            f"moto.ec2.models.{self.backend_delete}",
            side_effect=BotoServerError(400, "Mocked error"),
        ):
            resource_absent_result = self.salt_states[
                f"boto_vpc.{self.resource_type}_absent"
            ]("test")
            self.assertFalse(resource_absent_result["result"])
            self.assertTrue("Mocked error" in resource_absent_result["comment"])


@pytest.mark.skipif(HAS_BOTO is False, reason="The boto module must be installed.")
@pytest.mark.skipif(HAS_MOTO is False, reason="The moto module must be installed.")
@pytest.mark.skipif(
    _has_required_boto() is False,
    reason="The boto module must be greater than or equal to version {}".format(
        required_boto_version
    ),
)
class BotoVpcSubnetsTestCase(BotoVpcStateTestCaseBase, BotoVpcResourceTestCaseMixin):
    resource_type = "subnet"
    backend_create = "SubnetBackend.create_subnet"
    backend_delete = "SubnetBackend.delete_subnet"
    extra_kwargs = {"cidr_block": cidr_block}


@pytest.mark.skipif(HAS_BOTO is False, reason="The boto module must be installed.")
@pytest.mark.skipif(HAS_MOTO is False, reason="The moto module must be installed.")
@pytest.mark.skipif(
    _has_required_boto() is False,
    reason="The boto module must be greater than or equal to version {}".format(
        required_boto_version
    ),
)
class BotoVpcInternetGatewayTestCase(
    BotoVpcStateTestCaseBase, BotoVpcResourceTestCaseMixin
):
    resource_type = "internet_gateway"
    backend_create = "InternetGatewayBackend.create_internet_gateway"
    backend_delete = "InternetGatewayBackend.delete_internet_gateway"


@pytest.mark.skip(
    reason="Disabled for Python 3 due to upstream bugs: "
    "https://github.com/spulec/moto/issues/548 and "
    "https://github.com/gabrielfalcao/HTTPretty/issues/325",
)
@pytest.mark.skipif(HAS_BOTO is False, reason="The boto module must be installed.")
@pytest.mark.skipif(HAS_MOTO is False, reason="The moto module must be installed.")
@pytest.mark.skipif(
    _has_required_boto() is False,
    reason="The boto module must be greater than or equal to version {}".format(
        required_boto_version
    ),
)
class BotoVpcRouteTableTestCase(BotoVpcStateTestCaseBase, BotoVpcResourceTestCaseMixin):
    resource_type = "route_table"
    backend_create = "RouteTableBackend.create_route_table"
    backend_delete = "RouteTableBackend.delete_route_table"

    @mock_ec2_deprecated
    def test_present_with_subnets(self):
        vpc = self._create_vpc(name="test")
        subnet1 = self._create_subnet(
            vpc_id=vpc.id, cidr_block="10.0.0.0/25", name="test1"
        )
        subnet2 = self._create_subnet(
            vpc_id=vpc.id, cidr_block="10.0.0.128/25", name="test2"
        )

        route_table_present_result = self.salt_states["boto_vpc.route_table_present"](
            name="test",
            vpc_name="test",
            subnet_names=["test1"],
            subnet_ids=[subnet2.id],
        )

        associations = route_table_present_result["changes"]["new"][
            "subnets_associations"
        ]

        assoc_subnets = [x["subnet_id"] for x in associations]
        self.assertEqual(set(assoc_subnets), {subnet1.id, subnet2.id})

        route_table_present_result = self.salt_states["boto_vpc.route_table_present"](
            name="test", vpc_name="test", subnet_ids=[subnet2.id]
        )

        changes = route_table_present_result["changes"]

        old_subnets = [x["subnet_id"] for x in changes["old"]["subnets_associations"]]
        self.assertEqual(set(assoc_subnets), set(old_subnets))

        new_subnets = changes["new"]["subnets_associations"]
        self.assertEqual(new_subnets[0]["subnet_id"], subnet2.id)

    @mock_ec2_deprecated
    def test_present_with_routes(self):
        vpc = self._create_vpc(name="test")
        igw = self._create_internet_gateway(name="test", vpc_id=vpc.id)

        with patch.dict(botomod.__salt__, self.funcs):
            route_table_present_result = self.salt_states[
                "boto_vpc.route_table_present"
            ](
                name="test",
                vpc_name="test",
                routes=[
                    {"destination_cidr_block": "0.0.0.0/0", "gateway_id": igw.id},
                    {"destination_cidr_block": "10.0.0.0/24", "gateway_id": "local"},
                ],
            )
        routes = [
            x["gateway_id"]
            for x in route_table_present_result["changes"]["new"]["routes"]
        ]

        self.assertEqual(set(routes), {"local", igw.id})

        route_table_present_result = self.salt_states["boto_vpc.route_table_present"](
            name="test",
            vpc_name="test",
            routes=[{"destination_cidr_block": "10.0.0.0/24", "gateway_id": "local"}],
        )

        changes = route_table_present_result["changes"]

        old_routes = [x["gateway_id"] for x in changes["old"]["routes"]]
        self.assertEqual(set(routes), set(old_routes))

        self.assertEqual(changes["new"]["routes"][0]["gateway_id"], "local")



"""
Control Linux Containers via Salt

:depends: lxc execution module
"""

import copy
import logging
import os
import time
from collections import OrderedDict as _OrderedDict

import salt.client
import salt.key
import salt.utils.args
import salt.utils.cloud
import salt.utils.files
import salt.utils.stringutils
import salt.utils.virt

log = logging.getLogger(__name__)

# Don't shadow built-in's.
__func_alias__ = {"list_": "list"}


def _do(name, fun, path=None):
    """
    Invoke a function in the lxc module with no args

    path
        path to the container parent
        default: /var/lib/lxc (system default)

        .. versionadded:: 2015.8.0
    """
    host = find_guest(name, quiet=True, path=path)
    if not host:
        return False

    with salt.client.get_local_client(__opts__["conf_file"]) as client:
        cmd_ret = client.cmd_iter(
            host, f"lxc.{fun}", [name], kwarg={"path": path}, timeout=60
        )
        data = next(cmd_ret)
        data = data.get(host, {}).get("ret", None)
        if data:
            data = {host: data}
        return data


def _do_names(names, fun, path=None):
    """
    Invoke a function in the lxc module with no args

    path
        path to the container parent
        default: /var/lib/lxc (system default)

        .. versionadded:: 2015.8.0
    """
    ret = {}
    hosts = find_guests(names, path=path)
    if not hosts:
        return False

    with salt.client.get_local_client(__opts__["conf_file"]) as client:
        for host, sub_names in hosts.items():
            cmds = []
            for name in sub_names:
                cmds.append(
                    client.cmd_iter(
                        host,
                        f"lxc.{fun}",
                        [name],
                        kwarg={"path": path},
                        timeout=60,
                    )
                )
            for cmd in cmds:
                data = next(cmd)
                data = data.get(host, {}).get("ret", None)
                if data:
                    ret.update({host: data})
        return ret


def find_guest(name, quiet=False, path=None):
    """
    Returns the host for a container.

    path
        path to the container parent
        default: /var/lib/lxc (system default)

        .. versionadded:: 2015.8.0


    .. code-block:: bash

        salt-run lxc.find_guest name
    """
    if quiet:
        log.warning("'quiet' argument is being deprecated. Please migrate to --quiet")
    for data in _list_iter(path=path):
        host, l = next(iter(data.items()))
        for x in "running", "frozen", "stopped":
            if name in l[x]:
                if not quiet:
                    __jid_event__.fire_event(
                        {"data": host, "outputter": "lxc_find_host"}, "progress"
                    )
                return host
    return None


def find_guests(names, path=None):
    """
    Return a dict of hosts and named guests

    path
        path to the container parent
        default: /var/lib/lxc (system default)

        .. versionadded:: 2015.8.0

    """
    ret = {}
    names = names.split(",")
    for data in _list_iter(path=path):
        host, stat = next(iter(data.items()))
        for state in stat:
            for name in stat[state]:
                if name in names:
                    if host in ret:
                        ret[host].append(name)
                    else:
                        ret[host] = [name]
    return ret


def init(names, host=None, saltcloud_mode=False, quiet=False, **kwargs):
    """
    Initialize a new container


    .. code-block:: bash

        salt-run lxc.init name host=minion_id [cpuset=cgroups_cpuset] \\
                [cpushare=cgroups_cpushare] [memory=cgroups_memory] \\
                [template=lxc_template_name] [clone=original name] \\
                [profile=lxc_profile] [network_proflile=network_profile] \\
                [nic=network_profile] [nic_opts=nic_opts] \\
                [start=(true|false)] [seed=(true|false)] \\
                [install=(true|false)] [config=minion_config] \\
                [snapshot=(true|false)]

    names
        Name of the containers, supports a single name or a comma delimited
        list of names.

    host
        Minion on which to initialize the container **(required)**

    path
        path to the container parent
        default: /var/lib/lxc (system default)

        .. versionadded:: 2015.8.0

    saltcloud_mode
        init the container with the saltcloud opts format instead
        See lxc.init_interface module documentation

    cpuset
        cgroups cpuset.

    cpushare
        cgroups cpu shares.

    memory
        cgroups memory limit, in MB

        .. versionchanged:: 2015.5.0
            If no value is passed, no limit is set. In earlier Salt versions,
            not passing this value causes a 1024MB memory limit to be set, and
            it was necessary to pass ``memory=0`` to set no limit.

    template
        Name of LXC template on which to base this container

    clone
        Clone this container from an existing container

    profile
        A LXC profile (defined in config or pillar).

    network_profile
        Network profile to use for the container

        .. versionadded:: 2015.5.2

    nic
        .. deprecated:: 2015.5.0
            Use ``network_profile`` instead

    nic_opts
        Extra options for network interfaces. E.g.:

        ``{"eth0": {"mac": "aa:bb:cc:dd:ee:ff", "ipv4": "10.1.1.1", "ipv6": "2001:db8::ff00:42:8329"}}``

    start
        Start the newly created container.

    seed
        Seed the container with the minion config and autosign its key.
        Default: true

    install
        If salt-minion is not already installed, install it. Default: true

    config
        Optional config parameters. By default, the id is set to
        the name of the container.
    """
    path = kwargs.get("path", None)
    if quiet:
        log.warning("'quiet' argument is being deprecated. Please migrate to --quiet")
    ret = {"comment": "", "result": True}
    if host is None:
        # TODO: Support selection of host based on available memory/cpu/etc.
        ret["comment"] = "A host must be provided"
        ret["result"] = False
        return ret
    if isinstance(names, str):
        names = names.split(",")
    if not isinstance(names, list):
        ret["comment"] = "Container names are not formed as a list"
        ret["result"] = False
        return ret
    # check that the host is alive
    alive = False
    with salt.client.get_local_client(__opts__["conf_file"]) as client:
        try:
            if client.cmd(host, "test.ping", timeout=20).get(host, None):
                alive = True
        except (TypeError, KeyError):
            pass
        if not alive:
            ret["comment"] = f"Host {host} is not reachable"
            ret["result"] = False
            return ret

        log.info("Searching for LXC Hosts")
        data = __salt__["lxc.list"](host, quiet=True, path=path)
        for host, containers in data.items():
            for name in names:
                if name in sum(containers.values(), []):
                    log.info(
                        "Container '%s' already exists on host '%s', init "
                        "can be a NO-OP",
                        name,
                        host,
                    )
        if host not in data:
            ret["comment"] = f"Host '{host}' was not found"
            ret["result"] = False
            return ret

        kw = salt.utils.args.clean_kwargs(**kwargs)
        pub_key = kw.get("pub_key", None)
        priv_key = kw.get("priv_key", None)
        explicit_auth = pub_key and priv_key
        approve_key = kw.get("approve_key", True)
        seeds = {}
        seed_arg = kwargs.get("seed", True)
        if approve_key and not explicit_auth:
            with salt.key.Key(__opts__) as skey:
                all_minions = skey.all_keys().get("minions", [])
                for name in names:
                    seed = seed_arg
                    if name in all_minions:
                        try:
                            if client.cmd(name, "test.ping", timeout=20).get(
                                name, None
                            ):
                                seed = False
                        except (TypeError, KeyError):
                            pass
                    seeds[name] = seed
                    kv = salt.utils.virt.VirtKey(host, name, __opts__)
                    if kv.authorize():
                        log.info("Container key will be preauthorized")
                    else:
                        ret["comment"] = "Container key preauthorization failed"
                        ret["result"] = False
                        return ret

        log.info("Creating container(s) '%s' on host '%s'", names, host)

        cmds = []
        for name in names:
            args = [name]
            kw = salt.utils.args.clean_kwargs(**kwargs)
            if saltcloud_mode:
                kw = copy.deepcopy(kw)
                kw["name"] = name
                saved_kwargs = kw
                kw = client.cmd(
                    host,
                    "lxc.cloud_init_interface",
                    args + [kw],
                    tgt_type="list",
                    timeout=600,
                ).get(host, {})
                kw.update(saved_kwargs)
            name = kw.pop("name", name)
            # be sure not to seed an already seeded host
            kw["seed"] = seeds.get(name, seed_arg)
            if not kw["seed"]:
                kw.pop("seed_cmd", "")
            cmds.append(
                (
                    host,
                    name,
                    client.cmd_iter(host, "lxc.init", args, kwarg=kw, timeout=600),
                )
            )
        done = ret.setdefault("done", [])
        errors = ret.setdefault("errors", _OrderedDict())

        for ix, acmd in enumerate(cmds):
            hst, container_name, cmd = acmd
            containers = ret.setdefault(hst, [])
            herrs = errors.setdefault(hst, _OrderedDict())
            serrs = herrs.setdefault(container_name, [])
            sub_ret = next(cmd)
            error = None
            if isinstance(sub_ret, dict) and host in sub_ret:
                j_ret = sub_ret[hst]
                container = j_ret.get("ret", {})
                if container and isinstance(container, dict):
                    if not container.get("result", False):
                        error = container
                else:
                    error = "Invalid return for {}: {} {}".format(
                        container_name, container, sub_ret
                    )
            else:
                error = sub_ret
                if not error:
                    error = "unknown error (no return)"
            if error:
                ret["result"] = False
                serrs.append(error)
            else:
                container["container_name"] = name
                containers.append(container)
                done.append(container)

        # marking ping status as True only and only if we have at
        # least provisioned one container
        ret["ping_status"] = bool(len(done))

        # for all provisioned containers, last job is to verify
        # - the key status
        # - we can reach them
        for container in done:
            # explicitly check and update
            # the minion key/pair stored on the master
            container_name = container["container_name"]
            key = os.path.join(__opts__["pki_dir"], "minions", container_name)
            if explicit_auth:
                fcontent = ""
                if os.path.exists(key):
                    with salt.utils.files.fopen(key) as fic:
                        fcontent = salt.utils.stringutils.to_unicode(fic.read()).strip()
                pub_key = salt.utils.stringutils.to_unicode(pub_key)
                if pub_key.strip() != fcontent:
                    with salt.utils.files.fopen(key, "w") as fic:
                        fic.write(salt.utils.stringutils.to_str(pub_key))
                        fic.flush()
            mid = j_ret.get("mid", None)
            if not mid:
                continue

            def testping(**kw):
                mid_ = kw["mid"]
                ping = client.cmd(mid_, "test.ping", timeout=20)
                time.sleep(1)
                if ping:
                    return "OK"
                raise Exception(f"Unresponsive {mid_}")

            ping = salt.utils.cloud.wait_for_fun(testping, timeout=21, mid=mid)
            if ping != "OK":
                ret["ping_status"] = False
                ret["result"] = False

        # if no lxc detected as touched (either inited or verified)
        # we result to False
        if not done:
            ret["result"] = False
        if not quiet:
            __jid_event__.fire_event({"message": ret}, "progress")
        return ret


def cloud_init(names, host=None, quiet=False, **kwargs):
    """
    Wrapper for using lxc.init in saltcloud compatibility mode

    names
        Name of the containers, supports a single name or a comma delimited
        list of names.

    host
        Minion to start the container on. Required.

    path
        path to the container parent
        default: /var/lib/lxc (system default)

        .. versionadded:: 2015.8.0

    saltcloud_mode
        init the container with the saltcloud opts format instead
    """
    if quiet:
        log.warning("'quiet' argument is being deprecated. Please migrate to --quiet")
    return __salt__["lxc.init"](
        names=names, host=host, saltcloud_mode=True, quiet=quiet, **kwargs
    )


def _list_iter(host=None, path=None):
    """
    Return a generator iterating over hosts

    path
        path to the container parent
        default: /var/lib/lxc (system default)

        .. versionadded:: 2015.8.0
    """
    tgt = host or "*"
    with salt.client.get_local_client(__opts__["conf_file"]) as client:
        for container_info in client.cmd_iter(tgt, "lxc.list", kwarg={"path": path}):
            if not container_info:
                continue
            if not isinstance(container_info, dict):
                continue
            chunk = {}
            id_ = next(iter(container_info.keys()))
            if host and host != id_:
                continue
            if not isinstance(container_info[id_], dict):
                continue
            if "ret" not in container_info[id_]:
                continue
            if not isinstance(container_info[id_]["ret"], dict):
                continue
            chunk[id_] = container_info[id_]["ret"]
            yield chunk


def list_(host=None, quiet=False, path=None):
    """
    List defined containers (running, stopped, and frozen) for the named
    (or all) host(s).

    path
        path to the container parent
        default: /var/lib/lxc (system default)

        .. versionadded:: 2015.8.0

    .. code-block:: bash

        salt-run lxc.list [host=minion_id]
    """
    it = _list_iter(host, path=path)
    ret = {}
    for chunk in it:
        ret.update(chunk)
        if not quiet:
            __jid_event__.fire_event(
                {"data": chunk, "outputter": "lxc_list"}, "progress"
            )
    return ret


def purge(name, delete_key=True, quiet=False, path=None):
    """
    Purge the named container and delete its minion key if present.
    WARNING: Destroys all data associated with the container.

    path
        path to the container parent
        default: /var/lib/lxc (system default)

        .. versionadded:: 2015.8.0

    .. code-block:: bash

        salt-run lxc.purge name
    """
    data = _do_names(name, "destroy", path=path)
    if data is False:
        return data

    if delete_key:
        with salt.key.Key(__opts__) as skey:
            skey.delete_key(name)

    if data is None:
        return

    if not quiet:
        __jid_event__.fire_event({"data": data, "outputter": "lxc_purge"}, "progress")
    return data


def start(name, quiet=False, path=None):
    """
    Start the named container.

    path
        path to the container parent
        default: /var/lib/lxc (system default)

        .. versionadded:: 2015.8.0

    .. code-block:: bash

        salt-run lxc.start name
    """
    data = _do_names(name, "start", path=path)
    if data and not quiet:
        __jid_event__.fire_event({"data": data, "outputter": "lxc_start"}, "progress")
    return data


def stop(name, quiet=False, path=None):
    """
    Stop the named container.

    path
        path to the container parent
        default: /var/lib/lxc (system default)

        .. versionadded:: 2015.8.0

    .. code-block:: bash

        salt-run lxc.stop name
    """
    data = _do_names(name, "stop", path=path)
    if data and not quiet:
        __jid_event__.fire_event(
            {"data": data, "outputter": "lxc_force_off"}, "progress"
        )
    return data


def freeze(name, quiet=False, path=None):
    """
    Freeze the named container

    path
        path to the container parent
        default: /var/lib/lxc (system default)

        .. versionadded:: 2015.8.0

    .. code-block:: bash

        salt-run lxc.freeze name
    """
    data = _do_names(name, "freeze")
    if data and not quiet:
        __jid_event__.fire_event({"data": data, "outputter": "lxc_pause"}, "progress")
    return data


def unfreeze(name, quiet=False, path=None):
    """
    Unfreeze the named container

    path
        path to the container parent
        default: /var/lib/lxc (system default)

        .. versionadded:: 2015.8.0

    .. code-block:: bash

        salt-run lxc.unfreeze name
    """
    data = _do_names(name, "unfreeze", path=path)
    if data and not quiet:
        __jid_event__.fire_event({"data": data, "outputter": "lxc_resume"}, "progress")
    return data


def info(name, quiet=False, path=None):
    """
    Returns information about a container.

    path
        path to the container parent
        default: /var/lib/lxc (system default)

        .. versionadded:: 2015.8.0

    .. code-block:: bash

        salt-run lxc.info name
    """
    data = _do_names(name, "info", path=path)
    if data and not quiet:
        __jid_event__.fire_event({"data": data, "outputter": "lxc_info"}, "progress")
    return data



import pytest

pytestmark = [
    pytest.mark.skip_on_windows,
]


def test_salt_minion_ping(salt_cli, salt_minion, salt_master):
    """
    Test running a command against a targeted minion
    """
    assert salt_master.is_running()

    ret = salt_cli.run("test.ping", minion_tgt=salt_minion.id)
    assert ret.returncode == 0
    assert ret.data is True


def test_salt_minion_setproctitle(salt_cli, salt_minion, salt_master):
    """
    Test that setproctitle is working
    for the running Salt minion
    """
    assert salt_master.is_running()

    ret = salt_cli.run(
        "ps.pgrep", "MinionProcessManager", full=True, minion_tgt=salt_minion.id
    )
    assert ret.returncode == 0
    assert ret.data != ""



"""
Module for managing JBoss AS 7 through the CLI interface.

.. versionadded:: 2015.5.0

In order to run each function, jboss_config dictionary with the following properties must be passed:
 * cli_path: the path to jboss-cli script, for example: '/opt/jboss/jboss-7.0/bin/jboss-cli.sh'
 * controller: the IP address and port of controller, for example: 10.11.12.13:9999
 * cli_user: username to connect to jboss administration console if necessary
 * cli_password: password to connect to jboss administration console if necessary

Example:

.. code-block:: yaml

   jboss_config:
      cli_path: '/opt/jboss/jboss-7.0/bin/jboss-cli.sh'
      controller: 10.11.12.13:9999
      cli_user: 'jbossadm'
      cli_password: 'jbossadm'

"""

import logging
import re

import salt.utils.dictdiffer as dictdiffer
from salt.exceptions import SaltInvocationError

log = logging.getLogger(__name__)

__func_alias__ = {"reload_": "reload"}


def status(jboss_config, host=None, server_config=None):
    """
    Get status of running jboss instance.

    jboss_config
        Configuration dictionary with properties specified above.
    host
        The name of the host. JBoss domain mode only - and required if running in domain mode.
        The host name is the "name" attribute of the "host" element in host.xml
    server_config
        The name of the Server Configuration. JBoss Domain mode only - and required
        if running in domain mode.

    CLI Example:

    .. code-block:: bash

        salt '*' jboss7.status '{"cli_path": "integration.modules.sysmod.SysModuleTest.test_valid_docs", "controller": "10.11.12.13:9999", "cli_user": "jbossadm", "cli_password": "jbossadm"}'

    """
    log.debug("======================== MODULE FUNCTION: jboss7.status")
    if host is None and server_config is None:
        operation = ":read-attribute(name=server-state)"
    elif host is not None and server_config is not None:
        operation = '/host="{host}"/server-config="{server_config}"/:read-attribute(name=status)'.format(
            host=host, server_config=server_config
        )
    else:
        raise SaltInvocationError(
            "Invalid parameters. Must either pass both host and server_config or"
            " neither"
        )
    return __salt__["jboss7_cli.run_operation"](
        jboss_config, operation, fail_on_error=False, retries=0
    )


def stop_server(jboss_config, host=None):
    """
    Stop running jboss instance

    jboss_config
        Configuration dictionary with properties specified above.
    host
        The name of the host. JBoss domain mode only - and required if running in domain mode.
        The host name is the "name" attribute of the "host" element in host.xml

    CLI Example:

    .. code-block:: bash

        salt '*' jboss7.stop_server '{"cli_path": "integration.modules.sysmod.SysModuleTest.test_valid_docs", "controller": "10.11.12.13:9999", "cli_user": "jbossadm", "cli_password": "jbossadm"}'

    """
    log.debug("======================== MODULE FUNCTION: jboss7.stop_server")
    if host is None:
        operation = ":shutdown"
    else:
        operation = f'/host="{host}"/:shutdown'
    shutdown_result = __salt__["jboss7_cli.run_operation"](
        jboss_config, operation, fail_on_error=False
    )
    # JBoss seems to occasionaly close the channel immediately when :shutdown is sent
    if shutdown_result["success"] or (
        not shutdown_result["success"]
        and "Operation failed: Channel closed" in shutdown_result["stdout"]
    ):
        return shutdown_result
    else:
        raise Exception(
            """Cannot handle error, return code={retcode}, stdout='{stdout}', stderr='{stderr}' """.format(
                **shutdown_result
            )
        )


def reload_(jboss_config, host=None):
    """
    Reload running jboss instance

    jboss_config
        Configuration dictionary with properties specified above.
    host
        The name of the host. JBoss domain mode only - and required if running in domain mode.
        The host name is the "name" attribute of the "host" element in host.xml

    CLI Example:

    .. code-block:: bash

        salt '*' jboss7.reload '{"cli_path": "integration.modules.sysmod.SysModuleTest.test_valid_docs", "controller": "10.11.12.13:9999", "cli_user": "jbossadm", "cli_password": "jbossadm"}'

    """
    log.debug("======================== MODULE FUNCTION: jboss7.reload")
    if host is None:
        operation = ":reload"
    else:
        operation = f'/host="{host}"/:reload'
    reload_result = __salt__["jboss7_cli.run_operation"](
        jboss_config, operation, fail_on_error=False
    )
    # JBoss seems to occasionaly close the channel immediately when :reload is sent
    if reload_result["success"] or (
        not reload_result["success"]
        and (
            "Operation failed: Channel closed" in reload_result["stdout"]
            or "Communication error: java.util.concurrent.ExecutionException: Operation failed"
            in reload_result["stdout"]
        )
    ):
        return reload_result
    else:
        raise Exception(
            """Cannot handle error, return code={retcode}, stdout='{stdout}', stderr='{stderr}' """.format(
                **reload_result
            )
        )


def create_datasource(jboss_config, name, datasource_properties, profile=None):
    """
    Create datasource in running jboss instance

    jboss_config
        Configuration dictionary with properties specified above.
    name
        Datasource name
    datasource_properties
        A dictionary of datasource properties to be created:
          - driver-name: mysql
          - connection-url: 'jdbc:mysql://localhost:3306/sampleDatabase'
          - jndi-name: 'java:jboss/datasources/sampleDS'
          - user-name: sampleuser
          - password: secret
          - min-pool-size: 3
          - use-java-context: True
    profile
        The profile name (JBoss domain mode only)

    CLI Example:

    .. code-block:: bash

        salt '*' jboss7.create_datasource '{"cli_path": "integration.modules.sysmod.SysModuleTest.test_valid_docs", "controller": "10.11.12.13:9999", "cli_user": "jbossadm", "cli_password": "jbossadm"}' 'my_datasource' '{"driver-name": "mysql", "connection-url": "jdbc:mysql://localhost:3306/sampleDatabase", "jndi-name": "java:jboss/datasources/sampleDS", "user-name": "sampleuser", "password": "secret", "min-pool-size": 3, "use-java-context": True}'
    """
    log.debug(
        "======================== MODULE FUNCTION: jboss7.create_datasource, name=%s,"
        " profile=%s",
        name,
        profile,
    )
    ds_resource_description = __get_datasource_resource_description(
        jboss_config, name, profile
    )

    operation = '/subsystem=datasources/data-source="{name}":add({properties})'.format(
        name=name,
        properties=__get_properties_assignment_string(
            datasource_properties, ds_resource_description
        ),
    )
    if profile is not None:
        operation = f'/profile="{profile}"' + operation

    return __salt__["jboss7_cli.run_operation"](
        jboss_config, operation, fail_on_error=False
    )


def __get_properties_assignment_string(datasource_properties, ds_resource_description):
    assignment_strings = []
    ds_attributes = ds_resource_description["attributes"]
    for key, val in datasource_properties.items():
        assignment_strings.append(
            __get_single_assignment_string(key, val, ds_attributes)
        )

    return ",".join(assignment_strings)


def __get_single_assignment_string(key, val, ds_attributes):
    return f"{key}={__format_value(key, val, ds_attributes)}"


def __format_value(key, value, ds_attributes):
    type_ = ds_attributes[key]["type"]
    if type_ == "BOOLEAN":
        if value in ("true", "false"):
            return value
        elif isinstance(value, bool):
            if value:
                return "true"
            else:
                return "false"
        else:
            raise Exception(f"Don't know how to convert {value} to BOOLEAN type")

    elif type_ == "INT":
        return str(value)
    elif type_ == "STRING":
        return f'"{value}"'
    else:
        raise Exception(f"Don't know how to format value {value} of type {type_}")


def update_datasource(jboss_config, name, new_properties, profile=None):
    """
    Update an existing datasource in running jboss instance.
    If the property doesn't exist if will be created, if it does, it will be updated with the new value

    jboss_config
        Configuration dictionary with properties specified above.
    name
        Datasource name
    new_properties
        A dictionary of datasource properties to be updated. For example:
          - driver-name: mysql
          - connection-url: 'jdbc:mysql://localhost:3306/sampleDatabase'
          - jndi-name: 'java:jboss/datasources/sampleDS'
          - user-name: sampleuser
          - password: secret
          - min-pool-size: 3
          - use-java-context: True
    profile
        The profile name (JBoss domain mode only)

    CLI Example:

    .. code-block:: bash

        salt '*' jboss7.update_datasource '{"cli_path": "integration.modules.sysmod.SysModuleTest.test_valid_docs", "controller": "10.11.12.13:9999", "cli_user": "jbossadm", "cli_password": "jbossadm"}' 'my_datasource' '{"driver-name": "mysql", "connection-url": "jdbc:mysql://localhost:3306/sampleDatabase", "jndi-name": "java:jboss/datasources/sampleDS", "user-name": "sampleuser", "password": "secret", "min-pool-size": 3, "use-java-context": True}'

    """
    log.debug(
        "======================== MODULE FUNCTION: jboss7.update_datasource, name=%s,"
        " profile=%s",
        name,
        profile,
    )
    ds_result = __read_datasource(jboss_config, name, profile)
    current_properties = ds_result["result"]
    diff = dictdiffer.DictDiffer(new_properties, current_properties)
    changed_properties = diff.changed()

    ret = {"success": True, "comment": ""}
    if len(changed_properties) > 0:
        ds_resource_description = __get_datasource_resource_description(
            jboss_config, name, profile
        )
        ds_attributes = ds_resource_description["attributes"]
        for key in changed_properties:
            update_result = __update_datasource_property(
                jboss_config, name, key, new_properties[key], ds_attributes, profile
            )
            if not update_result["success"]:
                ret["result"] = False
                ret["comment"] = ret[
                    "comment"
                ] + "Could not update datasource property {} with value {},\n stdout: {}\n".format(
                    key, new_properties[key], update_result["stdout"]
                )

    return ret


def __get_datasource_resource_description(jboss_config, name, profile=None):
    log.debug(
        "======================== MODULE FUNCTION:"
        " jboss7.__get_datasource_resource_description, name=%s, profile=%s",
        name,
        profile,
    )

    operation = (
        '/subsystem=datasources/data-source="{name}":read-resource-description'.format(
            name=name
        )
    )
    if profile is not None:
        operation = f'/profile="{profile}"' + operation
    operation_result = __salt__["jboss7_cli.run_operation"](jboss_config, operation)
    if operation_result["outcome"]:
        return operation_result["result"]


def read_datasource(jboss_config, name, profile=None):
    """
    Read datasource properties in the running jboss instance.

    jboss_config
        Configuration dictionary with properties specified above.
    name
        Datasource name
    profile
        Profile name (JBoss domain mode only)

    CLI Example:

    .. code-block:: bash

        salt '*' jboss7.read_datasource '{"cli_path": "integration.modules.sysmod.SysModuleTest.test_valid_docs", "controller": "10.11.12.13:9999", "cli_user": "jbossadm", "cli_password": "jbossadm"}'
    """
    log.debug(
        "======================== MODULE FUNCTION: jboss7.read_datasource, name=%s",
        name,
    )
    return __read_datasource(jboss_config, name, profile)


def create_simple_binding(jboss_config, binding_name, value, profile=None):
    """
    Create a simple jndi binding in the running jboss instance

    jboss_config
        Configuration dictionary with properties specified above.
    binding_name
        Binding name to be created
    value
        Binding value
    profile
        The profile name (JBoss domain mode only)

    CLI Example:

    .. code-block:: bash

        salt '*' jboss7.create_simple_binding \\
                '{"cli_path": "integration.modules.sysmod.SysModuleTest.test_valid_docs", \\
                "controller": "10.11.12.13:9999", "cli_user": "jbossadm", "cli_password": "jbossadm"}' \\
                my_binding_name my_binding_value
       """
    log.debug(
        "======================== MODULE FUNCTION: jboss7.create_simple_binding,"
        " binding_name=%s, value=%s, profile=%s",
        binding_name,
        value,
        profile,
    )
    operation = (
        '/subsystem=naming/binding="{binding_name}":add(binding-type=simple,'
        ' value="{value}")'.format(
            binding_name=binding_name, value=__escape_binding_value(value)
        )
    )
    if profile is not None:
        operation = f'/profile="{profile}"' + operation
    return __salt__["jboss7_cli.run_operation"](jboss_config, operation)


def update_simple_binding(jboss_config, binding_name, value, profile=None):
    """
    Update the simple jndi binding in the running jboss instance

    jboss_config
        Configuration dictionary with properties specified above.
    binding_name
        Binding name to be updated
    value
        New binding value
    profile
        The profile name (JBoss domain mode only)

    CLI Example:

    .. code-block:: bash

        salt '*' jboss7.update_simple_binding '{"cli_path": "integration.modules.sysmod.SysModuleTest.test_valid_docs", "controller": "10.11.12.13:9999", "cli_user": "jbossadm", "cli_password": "jbossadm"}' my_binding_name my_binding_value
    """
    log.debug(
        "======================== MODULE FUNCTION: jboss7.update_simple_binding,"
        " binding_name=%s, value=%s, profile=%s",
        binding_name,
        value,
        profile,
    )
    operation = (
        '/subsystem=naming/binding="{binding_name}":write-attribute(name=value,'
        ' value="{value}")'.format(
            binding_name=binding_name, value=__escape_binding_value(value)
        )
    )
    if profile is not None:
        operation = f'/profile="{profile}"' + operation
    return __salt__["jboss7_cli.run_operation"](jboss_config, operation)


def read_simple_binding(jboss_config, binding_name, profile=None):
    """
    Read jndi binding in the running jboss instance

    jboss_config
        Configuration dictionary with properties specified above.
    binding_name
        Binding name to be created
    profile
        The profile name (JBoss domain mode only)

    CLI Example:

        .. code-block:: bash

        salt '*' jboss7.read_simple_binding '{"cli_path": "integration.modules.sysmod.SysModuleTest.test_valid_docs", "controller": "10.11.12.13:9999", "cli_user": "jbossadm", "cli_password": "jbossadm"}' my_binding_name
    """
    log.debug(
        "======================== MODULE FUNCTION: jboss7.read_simple_binding, %s",
        binding_name,
    )
    return __read_simple_binding(jboss_config, binding_name, profile=profile)


def __read_simple_binding(jboss_config, binding_name, profile=None):

    operation = '/subsystem=naming/binding="{binding_name}":read-resource'.format(
        binding_name=binding_name
    )
    if profile is not None:
        operation = f'/profile="{profile}"' + operation
    return __salt__["jboss7_cli.run_operation"](jboss_config, operation)


def __update_datasource_property(
    jboss_config, datasource_name, name, value, ds_attributes, profile=None
):
    log.debug(
        "======================== MODULE FUNCTION: jboss7.__update_datasource_property,"
        " datasource_name=%s, name=%s, value=%s, profile=%s",
        datasource_name,
        name,
        value,
        profile,
    )

    operation = '/subsystem=datasources/data-source="{datasource_name}":write-attribute(name="{name}",value={value})'.format(
        datasource_name=datasource_name,
        name=name,
        value=__format_value(name, value, ds_attributes),
    )
    if profile is not None:
        operation = f'/profile="{profile}"' + operation

    return __salt__["jboss7_cli.run_operation"](
        jboss_config, operation, fail_on_error=False
    )


def __read_datasource(jboss_config, name, profile=None):

    operation = '/subsystem=datasources/data-source="{name}":read-resource'.format(
        name=name
    )
    if profile is not None:
        operation = f'/profile="{profile}"' + operation

    operation_result = __salt__["jboss7_cli.run_operation"](jboss_config, operation)

    return operation_result


def __escape_binding_value(binding_name):
    result = binding_name.replace("\\", "\\\\\\\\")  # replace \ -> \\\\

    return result


def remove_datasource(jboss_config, name, profile=None):
    """
    Remove an existing datasource from the running jboss instance.

    jboss_config
        Configuration dictionary with properties specified above.
    name
        Datasource name
    profile
        The profile (JBoss domain mode only)

    CLI Example:

    .. code-block:: bash

        salt '*' jboss7.remove_datasource '{"cli_path": "integration.modules.sysmod.SysModuleTest.test_valid_docs", "controller": "10.11.12.13:9999", "cli_user": "jbossadm", "cli_password": "jbossadm"}' my_datasource_name
    """
    log.debug(
        "======================== MODULE FUNCTION: jboss7.remove_datasource, name=%s,"
        " profile=%s",
        name,
        profile,
    )

    operation = f"/subsystem=datasources/data-source={name}:remove"
    if profile is not None:
        operation = f'/profile="{profile}"' + operation

    return __salt__["jboss7_cli.run_operation"](
        jboss_config, operation, fail_on_error=False
    )


def deploy(jboss_config, source_file):
    """
    Deploy the application on the jboss instance from the local file system where minion is running.

    jboss_config
        Configuration dictionary with properties specified above.
    source_file
        Source file to deploy from

    CLI Example:

    .. code-block:: bash

        salt '*' jboss7.deploy '{"cli_path": "integration.modules.sysmod.SysModuleTest.test_valid_docs", "controller": "10.11.12.13:9999", "cli_user": "jbossadm", "cli_password": "jbossadm"}' /opt/deploy_files/my_deploy
    """
    log.debug(
        "======================== MODULE FUNCTION: jboss7.deploy, source_file=%s",
        source_file,
    )
    command = f"deploy {source_file} --force "
    return __salt__["jboss7_cli.run_command"](
        jboss_config, command, fail_on_error=False
    )


def list_deployments(jboss_config):
    """
    List all deployments on the jboss instance

    jboss_config
        Configuration dictionary with properties specified above.

     CLI Example:

     .. code-block:: bash

         salt '*' jboss7.list_deployments '{"cli_path": "integration.modules.sysmod.SysModuleTest.test_valid_docs", "controller": "10.11.12.13:9999", "cli_user": "jbossadm", "cli_password": "jbossadm"}'

    """
    log.debug("======================== MODULE FUNCTION: jboss7.list_deployments")
    command_result = __salt__["jboss7_cli.run_command"](jboss_config, "deploy")
    deployments = []
    if len(command_result["stdout"]) > 0:
        deployments = re.split("\\s*", command_result["stdout"])
    log.debug("deployments=%s", deployments)
    return deployments


def undeploy(jboss_config, deployment):
    """
    Undeploy the application from jboss instance

    jboss_config
        Configuration dictionary with properties specified above.
    deployment
        Deployment name to undeploy

    CLI Example:

    .. code-block:: bash

        salt '*' jboss7.undeploy '{"cli_path": "integration.modules.sysmod.SysModuleTest.test_valid_docs", "controller": "10.11.12.13:9999", "cli_user": "jbossadm", "cli_password": "jbossadm"}' my_deployment
    """
    log.debug(
        "======================== MODULE FUNCTION: jboss7.undeploy, deployment=%s",
        deployment,
    )
    command = f"undeploy {deployment} "
    return __salt__["jboss7_cli.run_command"](jboss_config, command)



"""
    salt.serializers.configparser
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    .. versionadded:: 2016.3.0

    Implements a configparser serializer.
"""

import configparser
import io

from salt.serializers import DeserializationError, SerializationError

__all__ = ["deserialize", "serialize", "available"]

available = True


def deserialize(stream_or_string, **options):
    """
    Deserialize any string or stream like object into a Python data structure.

    :param stream_or_string: stream or string to deserialize.
    :param options: options given to lower configparser module.
    """

    cp = configparser.ConfigParser(**options)

    try:
        if not isinstance(stream_or_string, (bytes, str)):
            cp.read_file(stream_or_string)
        else:
            cp.read_file(io.StringIO(stream_or_string))
        data = {}
        for section_name in cp.sections():
            section = {}
            for k, v in cp.items(section_name):
                section[k] = v
            data[section_name] = section
        return data
    except Exception as error:  # pylint: disable=broad-except
        raise DeserializationError(error)


def serialize(obj, **options):
    """
    Serialize Python data to a configparser formatted string or file.

    :param obj: the data structure to serialize
    :param options: options given to lower configparser module.
    """

    try:
        if not isinstance(obj, dict):
            raise TypeError(
                f"configparser can only serialize dictionaries, not {type(obj)}"
            )
        fp = options.pop("fp", None)
        cp = configparser.ConfigParser(**options)
        _read_dict(cp, obj)

        if fp:
            return cp.write(fp)
        else:
            s = io.StringIO()
            cp.write(s)
            return s.getvalue()
    except Exception as error:  # pylint: disable=broad-except
        raise SerializationError(error)


def _is_defaultsect(section_name):
    return section_name == configparser.DEFAULTSECT


def _read_dict(cp, dictionary):
    """
    Cribbed from python3's ConfigParser.read_dict function.
    """
    for section, keys in dictionary.items():
        section = str(section)
        if not _is_defaultsect(section):
            cp.add_section(section)

        for key, value in keys.items():
            key = cp.optionxform(str(key))
            if value is not None:
                value = str(value)
            cp.set(section, key, value)



"""
Tests for salt.loader.lazy
"""

import sys

import pytest

import salt.loader
import salt.loader.context
import salt.loader.lazy
import salt.utils.files


@pytest.fixture
def loader_dir(tmp_path):
    """
    Create a simple directory with a couple modules to load and run tests
    against.
    """
    mod_contents = """
    def __virtual__():
        return True

    def set_context(key, value):
        __context__[key] = value

    def get_context(key):
        return __context__[key]

    def get_opts(key):
        return __opts__.get(key, None)
    """
    with pytest.helpers.temp_file(
        "mod_a.py", directory=tmp_path, contents=mod_contents
    ), pytest.helpers.temp_file("mod_b.py", directory=tmp_path, contents=mod_contents):
        yield str(tmp_path)


def test_loaders_have_uniq_context(loader_dir):
    """
    Loaded functions run in the LazyLoader's context.
    """
    opts = {"optimization_order": [0, 1, 2]}
    loader_1 = salt.loader.lazy.LazyLoader([loader_dir], opts)
    loader_2 = salt.loader.lazy.LazyLoader([loader_dir], opts)
    loader_1._load_all()
    loader_2._load_all()
    assert loader_1.pack["__context__"] == {}
    assert loader_2.pack["__context__"] == {}
    loader_1["mod_a.set_context"]("foo", "bar")
    assert loader_1.pack["__context__"] == {"foo": "bar"}
    assert loader_1["mod_b.get_context"]("foo") == "bar"
    with pytest.raises(KeyError):
        loader_2["mod_a.get_context"]("foo")
    assert loader_2.pack["__context__"] == {}


def test_loaded_methods_are_loaded_func(loader_dir):
    """
    Functions loaded from LazyLoader's item lookups are LoadedFunc objects
    """
    opts = {"optimization_order": [0, 1, 2]}
    loader_1 = salt.loader.lazy.LazyLoader([loader_dir], opts)
    fun = loader_1["mod_a.get_context"]
    assert isinstance(fun, salt.loader.lazy.LoadedFunc)


def test_loaded_modules_are_loaded_mods(loader_dir):
    """
    Modules looked up as attributes of LazyLoaders are LoadedMod objects.
    """
    opts = {"optimization_order": [0, 1, 2]}
    loader_1 = salt.loader.lazy.LazyLoader([loader_dir], opts)
    mod = loader_1.mod_a
    assert isinstance(mod, salt.loader.lazy.LoadedMod)


def test_loaders_create_named_loader_contexts(loader_dir):
    """
    LazyLoader's create NamedLoaderContexts on the modules the load.
    """
    opts = {"optimization_order": [0, 1, 2]}
    loader_1 = salt.loader.lazy.LazyLoader([loader_dir], opts)
    mod = loader_1.mod_a
    assert isinstance(mod.mod, str)
    func = mod.set_context
    assert isinstance(func, salt.loader.lazy.LoadedFunc)
    module_name = func.func.__module__
    module = sys.modules[module_name]
    assert isinstance(module.__context__, salt.loader.context.NamedLoaderContext)
    wrapped_module_name = func.__module__
    wrapped_module = sys.modules[wrapped_module_name]
    assert isinstance(
        wrapped_module.__context__, salt.loader.context.NamedLoaderContext
    )
    assert module is wrapped_module


def test_loaders_convert_context_to_values(loader_dir):
    """
    LazyLoaders convert NamedLoaderContexts to values when instantiated.
    """
    loader_context = salt.loader.context.LoaderContext()
    grains_default = {
        "os": "linux",
    }
    grains = salt.loader.context.NamedLoaderContext(
        "grains", loader_context, grains_default
    )
    opts = {
        "optimization_order": [0, 1, 2],
        "grains": grains,
    }
    loader_1 = salt.loader.lazy.LazyLoader([loader_dir], opts)
    assert loader_1.opts["grains"] == grains_default
    # The loader's opts is a copy
    assert opts["grains"] == grains


def test_missing_loader_from_salt_internal_loaders():
    with pytest.raises(RuntimeError):
        salt.loader._module_dirs(
            {"extension_modules": "/tmp/foo"}, "missingmodules", "module"
        )


def test_loader_pack_always_has_opts(loader_dir):
    loader = salt.loader.lazy.LazyLoader([loader_dir], opts={"foo": "bar"})
    assert "__opts__" in loader.pack
    assert "foo" in loader.pack["__opts__"]
    assert loader.pack["__opts__"]["foo"] == "bar"


def test_loader_pack_opts_not_overwritten(loader_dir):
    opts = {"foo": "bar"}
    loader = salt.loader.lazy.LazyLoader(
        [loader_dir],
        opts={"foo": "bar"},
        pack={"__opts__": {"baz": "bif"}},
    )
    assert "__opts__" in loader.pack
    assert "foo" not in loader.pack["__opts__"]
    assert "baz" in loader.pack["__opts__"]
    assert loader.pack["__opts__"]["baz"] == "bif"


@pytest.mark.parametrize(
    "test_value, expected",
    [
        (True, True),
        (False, False),
        ("abc", True),
        (123, True),
    ],
)
def test_loaded_func_ensures_test_boolean(loader_dir, test_value, expected):
    """
    Functions loaded from LazyLoader's item lookups are LoadedFunc objects
    """
    opts = {"optimization_order": [0, 1, 2], "test": test_value}
    loader = salt.loader.lazy.LazyLoader([loader_dir], opts)
    loaded_fun = loader["mod_a.get_opts"]
    ret = loaded_fun("test")
    assert ret is expected



# pylint: disable=C0103,W0622
"""
Sphinx documentation for Salt
"""
import os
import pathlib
import re
import shutil
import sys
import textwrap
import time
import types

from sphinx.directives.other import TocTree
from sphinx.util import logging

log = logging.getLogger(__name__)

# -- Add paths to PYTHONPATH ---------------------------------------------------
try:
    docs_basepath = os.path.abspath(os.path.dirname(__file__))
except NameError:
    # sphinx-intl and six execute some code which will raise this NameError
    # assume we're in the doc/ directory
    docs_basepath = os.path.abspath(os.path.dirname("."))

addtl_paths = (
    os.pardir,  # salt itself (for autodoc)
    "_ext",  # custom Sphinx extensions
)

for addtl_path in addtl_paths:
    path = os.path.abspath(os.path.join(docs_basepath, addtl_path))
    sys.path.insert(0, path)

# We're now able to import salt
import salt.version  # isort:skip

formulas_dir = os.path.join(os.pardir, docs_basepath, "formulas")

# ----- Intersphinx Settings ------------------------------------------------>
intersphinx_mapping = {
    "python": (
        "https://docs.python.org/3",
        (
            "/usr/share/doc/python{}.{}/html/objects.inv".format(
                sys.version_info[0], sys.version_info[1]
            ),
            "/usr/share/doc/python/html/objects.inv",
            None,
        ),
    )
}
# <---- Intersphinx Settings -------------------------------------------------

# -- General Configuration -----------------------------------------------------

# Set a var if we're building docs for the live site or not
on_saltstack = "SALT_ON_SALTSTACK" in os.environ

project = "Salt"
# This is the default branch on GitHub for the Salt project
repo_primary_branch = "master"
if "LATEST_RELEASE" not in os.environ:
    salt_version = salt.version.__saltstack_version__
else:
    salt_version = salt.version.SaltStackVersion.parse(os.environ["LATEST_RELEASE"])

major_version = str(salt_version.major)
latest_release = ".".join([str(x) for x in salt_version.info])
previous_release = os.environ.get(
    "PREVIOUS_RELEASE", "previous_release"
)  # latest release from previous branch (3002.5)
previous_release_dir = os.environ.get(
    "PREVIOUS_RELEASE_DIR", "previous_release_dir"
)  # path on web server for previous branch (3002.5)
next_release = ""  # next release
next_release_dir = ""  # path on web server for next release branch

# Sphinx variable
# https://www.sphinx-doc.org/en/master/usage/configuration.html#confval-version
version = latest_release

today = ""
copyright = ""
if on_saltstack:
    today = "Generated on {} at {}.".format(
        time.strftime("%B %d, %Y"), time.strftime("%X %Z")
    )
    copyright = time.strftime("%Y")

# < --- START do not merge these settings to other branches START ---> #
build_type = os.environ.get(
    "BUILD_TYPE", repo_primary_branch
)  # latest, previous, master, next
# < --- END do not merge these settings to other branches END ---> #

# Set google custom search engine

if build_type == repo_primary_branch:
    release = latest_release
    search_cx = "011515552685726825874:v1had6i279q"  # master
    # search_cx = '011515552685726825874:x17j5zl74g8' # develop
elif build_type == "next":
    release = next_release
    search_cx = "011515552685726825874:ht0p8miksrm"  # latest
elif build_type == "previous":
    release = previous_release
    if release.startswith("3006"):
        search_cx = "2e4374de8af93a7b1"  # 3006
    elif release.startswith("3005"):
        search_cx = "57b1006b37edd9e79"  # 3005
    elif release.startswith("3004"):
        search_cx = "23cd7068705804111"  # 3004
    elif release.startswith("3003"):
        search_cx = "a70a1a73eef62aecd"  # 3003
    elif release.startswith("3002"):
        search_cx = "5026f4f2af0bdbe2d"  # 3002
    elif release.startswith("3001"):
        search_cx = "f0e4f298fa32b8a5e"  # 3001
    elif release.startswith("3000"):
        search_cx = "011515552685726825874:3skhaozjtyn"  # 3000
    elif release.startswith("2019.2"):
        search_cx = "011515552685726825874:huvjhlpptnm"  # 2019.2
    elif release.startswith("2018.3"):
        search_cx = "011515552685726825874:vadptdpvyyu"  # 2018.3
    elif release.startswith("2017.7"):
        search_cx = "011515552685726825874:w-hxmnbcpou"  # 2017.7
    elif release.startswith("2016.11"):
        search_cx = "011515552685726825874:dlsj745pvhq"  # 2016.11
    else:
        search_cx = "011515552685726825874:ht0p8miksrm"  # latest
else:  # latest or something else
    release = latest_release
    search_cx = "011515552685726825874:ht0p8miksrm"  # latest

needs_sphinx = "1.3"

spelling_lang = "en_US"
spelling_show_suggestions = True
language = "en"
locale_dirs = [
    "_locale",
]

master_doc = "contents"
templates_path = ["_templates"]
exclude_patterns = ["_build", "_incl/*", "ref/cli/_includes/*.rst"]

extensions = [
    "saltdomain",  # Must come early
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosummary",
    "sphinx.ext.extlinks",
    "sphinx.ext.imgconverter",
    "sphinx.ext.intersphinx",
    "sphinxcontrib.httpdomain",
    "saltrepo",
    "myst_parser",
    #'saltautodoc', # Must be AFTER autodoc
]

# Only enable spell-checking if enchant library is available
# This is optional for package builds but useful during development
try:
    import enchant

    extensions.append("sphinxcontrib.spelling")
except ImportError:
    log.info("Spell-checking disabled: enchant library not available")

modindex_common_prefix = ["salt."]

autosummary_generate = True
autosummary_generate_overwrite = False

# Smart dependency handling for documentation builds
# For man pages (lightweight CLI docs), we auto-mock missing dependencies
# For full HTML docs, we fail with helpful errors about what's missing
autodoc_mock_imports = []

# Detect if we're building man pages only (lightweight build)
# Man pages only document CLI tools and don't need full Salt imports
building_man_only = "man" in sys.argv and "html" not in sys.argv

# External dependencies that Salt imports at module level
# These need to be available for full HTML docs (autodoc) but can be mocked for man pages
_SALT_DEPENDENCIES = [
    "backports",  # Not available in Python 3.13+
    "distro",
    "jinja2",
    "looseversion",
    "msgpack",
    "packaging",
    "yaml",
]

_missing_deps = []
for dep in _SALT_DEPENDENCIES:
    try:
        __import__(dep)
    except ImportError:
        _missing_deps.append(dep)
        # Always mock missing dependencies to allow build to proceed
        autodoc_mock_imports.append(dep)

if _missing_deps:
    if building_man_only:
        # For man pages, this is expected - they don't need Salt modules
        log.info(
            "Building man pages with mocked dependencies: %s (this is normal for man pages)",
            ", ".join(_missing_deps),
        )
    else:
        # For HTML/full builds, warn that docs may be incomplete
        log.warning(
            "\n"
            "=" * 70 + "\n"
            "WARNING: Missing dependencies for full documentation build:\n"
            "  %s\n\n"
            "Autodoc will use mocked modules. Documentation will be generated but\n"
            "may be incomplete or show incorrect type hints.\n\n"
            "For complete documentation, install with:\n"
            "  pip install %s\n" + "=" * 70,
            ", ".join(_missing_deps),
            " ".join(_missing_deps),
        )

# strip git rev as there won't necessarily be a release based on it
stripped_release = re.sub(r"-\d+-g[0-9a-f]+$", "", release)

# Define a substitution for linking to the latest release tarball
rst_prolog = """\
.. |current_release_doc| replace:: :doc:`/topics/releases/{release}`
.. |saltrepo| replace:: https://github.com/saltstack/salt
.. _`salt-users`: https://groups.google.com/forum/#!forum/salt-users
.. _`salt-announce`: https://groups.google.com/forum/#!forum/salt-announce
.. _`salt-packagers`: https://groups.google.com/forum/#!forum/salt-packagers
.. _`salt-discord`: https://discord.com/invite/J7b7EscrAs

""".format(
    release=stripped_release
)

# A shortcut for linking to tickets on the GitHub issue tracker
extlinks = {
    "blob": (
        "https://github.com/saltstack/salt/blob/%s/%%s" % repo_primary_branch,
        "%s",
    ),
    "issue": ("https://github.com/saltstack/salt/issues/%s", "issue %s"),
    "pull": ("https://github.com/saltstack/salt/pull/%s", "PR %s"),
    "formula_url": ("https://github.com/saltstack-formulas/%s", "url %s"),
}

myst_gfm_only = True

# ----- Localization -------------------------------------------------------->
locale_dirs = ["locale/"]
gettext_compact = False
# <---- Localization ---------------------------------------------------------


### HTML options
# set 'HTML_THEME=saltstack' to use previous theme
html_theme = os.environ.get("HTML_THEME", "saltstack2")
html_theme_path = ["_themes"]
html_title = ""
html_short_title = "Salt"

html_static_path = ["_static"]
html_logo = None  # specified in the theme layout.html
html_favicon = "favicon.ico"
smartquotes = False

# Use Google customized search or use Sphinx built-in JavaScript search
if on_saltstack:
    html_search_template = "googlesearch.html"
else:
    html_search_template = "searchbox.html"

html_additional_pages = {
    "404": "404.html",
}

html_default_sidebars = [
    html_search_template,
    "version.html",
    "localtoc.html",
    "relations.html",
    "sourcelink.html",
    "saltstack.html",
]
html_sidebars = {
    "ref/**/all/salt.*": [
        html_search_template,
        "version.html",
        "modules-sidebar.html",
        "localtoc.html",
        "relations.html",
        "sourcelink.html",
        "saltstack.html",
    ],
    "ref/formula/all/*": [],
}

html_context = {
    "on_saltstack": on_saltstack,
    "html_default_sidebars": html_default_sidebars,
    "github_base": "https://github.com/saltstack/salt",
    "github_issues": "https://github.com/saltstack/salt/issues",
    "github_downloads": "https://github.com/saltstack/salt/downloads",
    "latest_release": latest_release,
    "previous_release": previous_release,
    "previous_release_dir": previous_release_dir,
    "next_release": next_release,
    "next_release_dir": next_release_dir,
    "search_cx": search_cx,
    "build_type": build_type,
    "today": today,
    "copyright": copyright,
    "repo_primary_branch": repo_primary_branch,
}

html_use_index = True
html_last_updated_fmt = "%b %d, %Y"
html_show_sourcelink = False
html_show_sphinx = True
html_show_copyright = True

### Latex options

latex_documents = [
    ("contents", "Salt.tex", "Salt Documentation", "VMware, Inc.", "manual"),
]

latex_logo = "_static/salt-logo.png"

latex_elements = {
    "inputenc": "",  # use XeTeX instead of the inputenc LaTeX package.
    "utf8extra": "",
    "preamble": r"""
    \usepackage{fontspec}
    \setsansfont{Linux Biolinum O}
    \setromanfont{Linux Libertine O}
    \setmonofont{Source Code Pro}
""",
}
### Linux Biolinum, Linux Libertine: http://www.linuxlibertine.org/
### Source Code Pro: https://github.com/adobe-fonts/source-code-pro/releases


### Linkcheck options
linkcheck_ignore = [
    r"http://127.0.0.1",
    r"http://salt:\d+",
    r"http://local:\d+",
    r"https://console.aws.amazon.com",
    r"http://192.168.33.10",
    r"http://domain:\d+",
    r"http://123.456.789.012:\d+",
    r"http://localhost",
    r"https://groups.google.com/forum/#!forum/salt-users",
    r"https://www.elastic.co/logstash/docs/latest/inputs/udp",
    r"https://www.elastic.co/logstash/docs/latest/inputs/zeromq",
    r"http://www.youtube.com/saltstack",
    r"https://raven.readthedocs.io",
    r"https://getsentry.com",
    r"https://salt-cloud.readthedocs.io",
    r"https://salt.readthedocs.io",
    r"http://www.pip-installer.org/",
    r"http://www.windowsazure.com/",
    r"https://github.com/watching",
    r"dash-feed://",
    r"https://github.com/saltstack/salt/",
    r"https://bootstrap.saltproject.io",
    r"https://raw.githubusercontent.com/saltstack/salt-bootstrap/stable/bootstrap-salt.sh",
    r"media.readthedocs.org/dash/salt/latest/salt.xml",
    r"https://portal.aws.amazon.com/gp/aws/securityCredentials",
    r"dash-feed://https%3A//media.readthedocs.org/dash/salt/latest/salt.xml",
    r"(?i)dns:.*",
    r"TCP:4506",
    r"https?://",
    r"https://cloud.github.com/downloads/saltstack/.*",
    r"https://INFOBLOX/.*",
    r"https://SOMESERVERIP:.*",
    r"https://community.saltstack.com/.*",
    # GitHub Users
    r"https://github.com/[^/]$",
    # GitHub Salt Forks
    r"https://github.com/[^/]/salt$",
    r"tag:key=value",
    r"jdbc:mysql:.*",
    r"http:post",
]
linkcheck_exclude_documents = [
    r"topics/releases/(2015|2016)\..*\.rst",
    r"topics/releases/saltapi/0\.8\.0.*",
]
linkcheck_timeout = 10
linkcheck_anchors = False

### Manpage options
# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
authors = [
    "Thomas S. Hatch <thatch45@gmail.com> and many others, please see the Authors file",
]

man_pages = [
    ("ref/cli/salt", "salt", "salt", authors, 1),
    ("ref/cli/salt-master", "salt-master", "salt-master Documentation", authors, 1),
    ("ref/cli/salt-minion", "salt-minion", "salt-minion Documentation", authors, 1),
    ("ref/cli/salt-key", "salt-key", "salt-key Documentation", authors, 1),
    ("ref/cli/salt-cp", "salt-cp", "salt-cp Documentation", authors, 1),
    ("ref/cli/salt-call", "salt-call", "salt-call Documentation", authors, 1),
    ("ref/cli/salt-proxy", "salt-proxy", "salt-proxy Documentation", authors, 1),
    ("ref/cli/salt-syndic", "salt-syndic", "salt-syndic Documentation", authors, 1),
    ("ref/cli/salt-run", "salt-run", "salt-run Documentation", authors, 1),
    ("ref/cli/salt-ssh", "salt-ssh", "salt-ssh Documentation", authors, 1),
    ("ref/cli/salt-cloud", "salt-cloud", "Salt Cloud Command", authors, 1),
    ("ref/cli/salt-api", "salt-api", "salt-api Command", authors, 1),
    ("ref/cli/spm", "spm", "Salt Package Manager Command", authors, 1),
]


def skip_mod_init_member(app, what, name, obj, skip, options):
    # pylint: disable=too-many-arguments,unused-argument
    if name.startswith("_"):
        return True
    if isinstance(obj, types.FunctionType) and obj.__name__ == "mod_init":
        return True
    return False


def _normalize_version(args):
    _, path = args
    return ".".join([x.zfill(4) for x in (path.split("/")[-1].split("."))])


class ReleasesTree(TocTree):
    option_spec = dict(TocTree.option_spec)

    def run(self):
        rst = super().run()
        entries = rst[0][0]["entries"][:]
        entries.sort(key=_normalize_version, reverse=True)
        rst[0][0]["entries"][:] = entries
        return rst


def copy_release_templates_pre(app):
    app._copied_release_files = []
    docs_path = pathlib.Path(docs_basepath)
    release_files_dir = docs_path / "topics" / "releases"
    release_template_files_dir = release_files_dir / "templates"
    for fpath in release_template_files_dir.iterdir():
        dest = release_files_dir / fpath.name.replace(".template", "")
        if dest.exists():
            continue
        log.info(
            "Copying '%s' -> '%s' just for this build ...",
            fpath.relative_to(docs_path),
            dest.relative_to(docs_path),
        )
        app._copied_release_files.append(dest)
        shutil.copyfile(fpath, dest)


def copy_release_templates_post(app, exception):
    docs_path = pathlib.Path(docs_basepath)
    for fpath in app._copied_release_files:
        log.info(
            "The release file '%s' was copied for the build, but its not in "
            "version control system. Deleting.",
            fpath.relative_to(docs_path),
        )
        fpath.unlink()


def extract_module_deprecations(app, what, name, obj, options, lines):
    """
    Add a warning to the modules being deprecated into extensions.
    """
    # https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#event-autodoc-process-docstring
    if what != "module":
        # We're only interested in module deprecations
        return

    try:
        deprecated_info = obj.__deprecated__
    except AttributeError:
        # The module is not deprecated
        return

    _version, _extension, _url = deprecated_info
    msg = textwrap.dedent(
        f"""
        .. warning::

            This module will be removed from Salt in version {_version} in favor of
            the `{_extension} Salt Extension <{_url}>`_.

        """
    )
    # Modify the docstring lines in-place
    lines[:] = msg.splitlines() + lines


def setup(app):
    app.add_directive("releasestree", ReleasesTree)
    app.connect("autodoc-skip-member", skip_mod_init_member)
    app.connect("builder-inited", copy_release_templates_pre)
    app.connect("build-finished", copy_release_templates_post)
    app.connect("autodoc-process-docstring", extract_module_deprecations)



import logging
import multiprocessing
import signal
import subprocess
import sys
import time

import pytest
from pytestshellutils.utils.processes import terminate_process

from salt._logging.handlers import DeferredStreamHandler
from salt._logging.impl import set_lowest_log_level
from salt.utils.nb_popen import NonBlockingPopen
from tests.support.helpers import CaptureOutput, dedent
from tests.support.runtests import RUNTIME_VARS

log = logging.getLogger(__name__)


def _sync_with_handlers_proc_target():

    with CaptureOutput() as stds:
        handler = DeferredStreamHandler(sys.stderr)
        set_lowest_log_level(logging.DEBUG)
        formatter = logging.Formatter("%(message)s")
        handler.setFormatter(formatter)
        logging.root.addHandler(handler)
        logger = logging.getLogger(__name__)
        logger.info("Foo")
        logger.info("Bar")
        logging.root.removeHandler(handler)

        assert not stds.stdout
        assert not stds.stderr

        stream_handler = logging.StreamHandler(sys.stderr)

        # Sync with the other handlers
        handler.sync_with_handlers([stream_handler])

        assert not stds.stdout
        assert stds.stderr == "Foo\nBar\n"


def _deferred_write_on_flush_proc_target():

    with CaptureOutput() as stds:
        handler = DeferredStreamHandler(sys.stderr)
        set_lowest_log_level(logging.DEBUG)
        formatter = logging.Formatter("%(message)s")
        handler.setFormatter(formatter)
        logging.root.addHandler(handler)
        logger = logging.getLogger(__name__)
        logger.info("Foo")
        logger.info("Bar")
        logging.root.removeHandler(handler)

        assert not stds.stdout
        assert not stds.stderr

        # Flush the handler
        handler.flush()
        assert not stds.stdout
        assert stds.stderr == "Foo\nBar\n"


def test_sync_with_handlers():
    proc = multiprocessing.Process(target=_sync_with_handlers_proc_target)
    proc.start()
    proc.join()
    assert proc.exitcode == 0


def test_deferred_write_on_flush():
    proc = multiprocessing.Process(target=_deferred_write_on_flush_proc_target)
    proc.start()
    proc.join()
    assert proc.exitcode == 0


def test_deferred_write_on_atexit(tmp_path):
    # Python will .flush() and .close() all logging handlers at interpreter shutdown.
    # This should be enough to flush our deferred messages.
    pyscript = dedent(
        r"""
        import sys
        import time
        import logging

        CODE_DIR = {!r}
        if CODE_DIR in sys.path:
            sys.path.remove(CODE_DIR)
        sys.path.insert(0, CODE_DIR)

        from salt._logging.handlers import DeferredStreamHandler
        # Reset any logging handlers we might have already
        logging.root.handlers[:] = []

        handler = DeferredStreamHandler(sys.stderr)
        handler.setLevel(logging.DEBUG)
        logging.root.addHandler(handler)

        log = logging.getLogger(__name__)
        sys.stdout.write('STARTED\n')
        sys.stdout.flush()
        log.debug('Foo')
        sys.exit(0)
    """.format(
            RUNTIME_VARS.CODE_DIR
        )
    )
    script_path = tmp_path / "atexit_deferred_logging_test.py"
    script_path.write_text(pyscript, encoding="utf-8")

    proc = NonBlockingPopen(
        [sys.executable, str(script_path)],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    out = b""
    err = b""

    # This test should never take more than 5 seconds
    execution_time = 5
    max_time = time.time() + execution_time
    try:
        # Just loop consuming output
        while True:
            if time.time() > max_time:
                pytest.fail(f"Script didn't exit after {execution_time} second")

            time.sleep(0.125)
            _out = proc.recv()
            _err = proc.recv_err()
            if _out:
                out += _out
            if _err:
                err += _err

            if _out is None and _err is None:
                # The script exited
                break

            if proc.poll() is not None:
                # The script exited
                break
    finally:
        terminate_process(proc.pid, kill_children=True)
    if b"Foo" not in err:
        pytest.fail(f"'Foo' should be in stderr and it's not: {err}")


@pytest.mark.skip_on_windows(reason="Windows does not support SIGINT")
def test_deferred_write_on_sigint(tmp_path):
    pyscript = dedent(
        r"""
        import sys
        import time
        import signal
        import logging

        CODE_DIR = {!r}
        if CODE_DIR in sys.path:
            sys.path.remove(CODE_DIR)
        sys.path.insert(0, CODE_DIR)

        from salt._logging.handlers import DeferredStreamHandler
        # Reset any logging handlers we might have already
        logging.root.handlers[:] = []

        handler = DeferredStreamHandler(sys.stderr)
        handler.setLevel(logging.DEBUG)
        logging.root.addHandler(handler)

        if signal.getsignal(signal.SIGINT) != signal.default_int_handler:
            # Looking at you Debian based distros :/
            signal.signal(signal.SIGINT, signal.default_int_handler)

        log = logging.getLogger(__name__)

        start_printed = False
        while True:
            try:
                log.debug('Foo')
                if start_printed is False:
                    sys.stdout.write('STARTED\n')
                    sys.stdout.write('SIGINT HANDLER: {{!r}}\n'.format(signal.getsignal(signal.SIGINT)))
                    sys.stdout.flush()
                    start_printed = True
                time.sleep(0.125)
            except (KeyboardInterrupt, SystemExit):
                log.info('KeyboardInterrupt caught')
                sys.stdout.write('KeyboardInterrupt caught\n')
                sys.stdout.flush()
                break
        log.info('EXITING')
        sys.stdout.write('EXITING\n')
        sys.stdout.flush()
        sys.exit(0)
        """.format(
            RUNTIME_VARS.CODE_DIR
        )
    )
    script_path = tmp_path / "sigint_deferred_logging_test.py"
    script_path.write_text(pyscript, encoding="utf-8")

    proc = NonBlockingPopen(
        [sys.executable, str(script_path)],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    out = b""
    err = b""

    # Test should take less than 20 seconds, way less
    execution_time = 10
    start = time.time()
    max_time = time.time() + execution_time
    try:
        signalled = False
        log.info("Starting Loop")
        while True:

            time.sleep(0.125)
            _out = proc.recv()
            _err = proc.recv_err()
            if _out:
                out += _out
            if _err:
                err += _err

            if b"STARTED" in out and not signalled:
                # Enough time has passed
                proc.send_signal(signal.SIGINT)
                signalled = True
                log.debug("Sent SIGINT after: %s", time.time() - start)

            if signalled is False:
                if out:
                    pytest.fail(
                        "We have stdout output when there should be none: {}".format(
                            out
                        )
                    )
                if err:
                    pytest.fail(
                        "We have stderr output when there should be none: {}".format(
                            err
                        )
                    )

            if _out is None and _err is None:
                log.info("_out and _err are None")
                if b"Foo" not in err:
                    pytest.fail(
                        "No more output and 'Foo' should be in stderr and it's not: {}".format(
                            err
                        )
                    )
                break

            if proc.poll() is not None:
                log.debug("poll() is not None")
                if b"Foo" not in err:
                    pytest.fail(
                        "Process terminated and 'Foo' should be in stderr and it's not: {}".format(
                            err
                        )
                    )
                break

            if time.time() > max_time:
                log.debug("Reached max time")
                if b"Foo" not in err:
                    pytest.fail(
                        "'Foo' should be in stderr and it's not:\n{0}\nSTDERR:\n{0}\n{1}\n{0}\nSTDOUT:\n{0}\n{2}\n{0}".format(
                            "-" * 80, err, out
                        )
                    )
    finally:
        terminate_process(proc.pid, kill_children=True)
    log.debug("Test took %s seconds", time.time() - start)



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.modules.keyboard as keyboard
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {keyboard: {}}


# 'get_sys' function tests: 1


def test_get_sys():
    """
    Test if it get current system keyboard setting
    """
    mock = MagicMock(return_value="X11 Layout=us")
    with patch.dict(keyboard.__grains__, {"os_family": "RedHat"}):
        with patch.dict(keyboard.__salt__, {"cmd.run": mock}):
            assert keyboard.get_sys() == "us"


# 'set_sys' function tests: 1


def test_set_sys():
    """
    Test if it set current system keyboard setting
    """
    mock = MagicMock(return_value="us")
    with patch.dict(keyboard.__grains__, {"os_family": "RedHat"}):
        with patch.dict(keyboard.__salt__, {"cmd.run": mock}):
            with patch.dict(keyboard.__salt__, {"file.sed": MagicMock()}):
                assert keyboard.set_sys("us") == "us"


# 'get_x' function tests: 1


def test_get_x():
    """
    Test if it get current X keyboard setting
    """
    mock = MagicMock(return_value="layout:     us")
    with patch.dict(keyboard.__salt__, {"cmd.run": mock}):
        assert keyboard.get_x() == "us"


# 'set_x' function tests: 1


def test_set_x():
    """
    Test if it set current X keyboard setting
    """
    mock = MagicMock(return_value="us")
    with patch.dict(keyboard.__salt__, {"cmd.run": mock}):
        assert keyboard.set_x("us") == "us"



import salt.utils.minions
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

NODEGROUPS = {
    "group1": "L@host1,host2,host3",
    "group2": ["G@foo:bar", "or", "web1*"],
    "group3": ["N@group1", "or", "N@group2"],
    "group4": ["host4", "host5", "host6"],
    "group5": "N@group4",
    "group6": "N@group3",
    "group7": ["host1"],
}

EXPECTED = {
    "group1": ["L@host1,host2,host3"],
    "group2": ["G@foo:bar", "or", "web1*"],
    "group3": [
        "(",
        "(",
        "L@host1,host2,host3",
        ")",
        "or",
        "(",
        "G@foo:bar",
        "or",
        "web1*",
        ")",
        ")",
    ],
    "group4": ["L@host4,host5,host6"],
    "group5": ["(", "L@host4,host5,host6", ")"],
    "group6": [
        "(",
        "(",
        "(",
        "L@host1,host2,host3",
        ")",
        "or",
        "(",
        "G@foo:bar",
        "or",
        "web1*",
        ")",
        ")",
        ")",
    ],
    "group7": ["L@host1"],
}


class MinionsTestCase(TestCase):
    """
    TestCase for salt.utils.minions module functions
    """

    def test_nodegroup_comp(self):
        """
        Test a simple string nodegroup
        """
        for nodegroup in NODEGROUPS:
            expected = EXPECTED[nodegroup]
            ret = salt.utils.minions.nodegroup_comp(nodegroup, NODEGROUPS)
            self.assertEqual(ret, expected)


class CkMinionsTestCase(TestCase):
    """
    TestCase for salt.utils.minions.CkMinions class
    """

    def setUp(self):
        self.ckminions = salt.utils.minions.CkMinions({"minion_data_cache": True})

    def test_spec_check(self):
        # Test spec-only rule
        auth_list = ["@runner"]
        ret = self.ckminions.spec_check(auth_list, "test.arg", {}, "runner")
        self.assertTrue(ret)
        ret = self.ckminions.spec_check(auth_list, "test.arg", {}, "wheel")
        self.assertFalse(ret)
        ret = self.ckminions.spec_check(auth_list, "testarg", {}, "runner")
        mock_ret = {
            "error": {
                "name": "SaltInvocationError",
                "message": "A command invocation error occurred: Check syntax.",
            }
        }
        self.assertDictEqual(mock_ret, ret)

        # Test spec in plural form
        auth_list = ["@runners"]
        ret = self.ckminions.spec_check(auth_list, "test.arg", {}, "runner")
        self.assertTrue(ret)
        ret = self.ckminions.spec_check(auth_list, "test.arg", {}, "wheel")
        self.assertFalse(ret)

        # Test spec with module.function restriction
        auth_list = [{"@runner": "test.arg"}]
        ret = self.ckminions.spec_check(auth_list, "test.arg", {}, "runner")
        self.assertTrue(ret)
        ret = self.ckminions.spec_check(auth_list, "test.arg", {}, "wheel")
        self.assertFalse(ret)
        ret = self.ckminions.spec_check(auth_list, "tes.arg", {}, "runner")
        self.assertFalse(ret)
        ret = self.ckminions.spec_check(auth_list, "test.ar", {}, "runner")
        self.assertFalse(ret)

        # Test function name is a regex
        auth_list = [{"@runner": "test.arg.*some"}]
        ret = self.ckminions.spec_check(auth_list, "test.arg", {}, "runner")
        self.assertFalse(ret)
        ret = self.ckminions.spec_check(auth_list, "test.argsome", {}, "runner")
        self.assertTrue(ret)
        ret = self.ckminions.spec_check(auth_list, "test.arg_aaa_some", {}, "runner")
        self.assertTrue(ret)

        # Test a list of funcs
        auth_list = [{"@runner": ["test.arg", "jobs.active"]}]
        ret = self.ckminions.spec_check(auth_list, "test.arg", {}, "runner")
        self.assertTrue(ret)
        ret = self.ckminions.spec_check(auth_list, "jobs.active", {}, "runner")
        self.assertTrue(ret)
        ret = self.ckminions.spec_check(auth_list, "test.active", {}, "runner")
        self.assertFalse(ret)
        ret = self.ckminions.spec_check(auth_list, "jobs.arg", {}, "runner")
        self.assertFalse(ret)

        # Test args-kwargs rules
        auth_list = [
            {
                "@runner": {
                    "test.arg": {
                        "args": ["1", "2"],
                        "kwargs": {"aaa": "bbb", "ccc": "ddd"},
                    }
                }
            }
        ]
        ret = self.ckminions.spec_check(auth_list, "test.arg", {}, "runner")
        self.assertFalse(ret)
        args = {"arg": ["1", "2"], "kwarg": {"aaa": "bbb", "ccc": "ddd"}}
        ret = self.ckminions.spec_check(auth_list, "test.arg", args, "runner")
        self.assertTrue(ret)
        args = {"arg": ["1", "2", "3"], "kwarg": {"aaa": "bbb", "ccc": "ddd"}}
        ret = self.ckminions.spec_check(auth_list, "test.arg", args, "runner")
        self.assertTrue(ret)
        args = {"arg": ["1", "2"], "kwarg": {"aaa": "bbb", "ccc": "ddd", "zzz": "zzz"}}
        ret = self.ckminions.spec_check(auth_list, "test.arg", args, "runner")
        self.assertTrue(ret)
        args = {"arg": ["1", "2"], "kwarg": {"aaa": "bbb", "ccc": "ddc"}}
        ret = self.ckminions.spec_check(auth_list, "test.arg", args, "runner")
        self.assertFalse(ret)
        args = {"arg": ["1", "2"], "kwarg": {"aaa": "bbb"}}
        ret = self.ckminions.spec_check(auth_list, "test.arg", args, "runner")
        self.assertFalse(ret)
        args = {"arg": ["1", "3"], "kwarg": {"aaa": "bbb", "ccc": "ddd"}}
        ret = self.ckminions.spec_check(auth_list, "test.arg", args, "runner")
        self.assertFalse(ret)
        args = {"arg": ["1"], "kwarg": {"aaa": "bbb", "ccc": "ddd"}}
        ret = self.ckminions.spec_check(auth_list, "test.arg", args, "runner")
        self.assertFalse(ret)
        args = {"kwarg": {"aaa": "bbb", "ccc": "ddd"}}
        ret = self.ckminions.spec_check(auth_list, "test.arg", args, "runner")
        self.assertFalse(ret)
        args = {
            "arg": ["1", "2"],
        }
        ret = self.ckminions.spec_check(auth_list, "test.arg", args, "runner")
        self.assertFalse(ret)

        # Test kwargs only
        auth_list = [
            {"@runner": {"test.arg": {"kwargs": {"aaa": "bbb", "ccc": "ddd"}}}}
        ]
        ret = self.ckminions.spec_check(auth_list, "test.arg", {}, "runner")
        self.assertFalse(ret)
        args = {"arg": ["1", "2"], "kwarg": {"aaa": "bbb", "ccc": "ddd"}}
        ret = self.ckminions.spec_check(auth_list, "test.arg", args, "runner")
        self.assertTrue(ret)

        # Test args only
        auth_list = [{"@runner": {"test.arg": {"args": ["1", "2"]}}}]
        ret = self.ckminions.spec_check(auth_list, "test.arg", {}, "runner")
        self.assertFalse(ret)
        args = {"arg": ["1", "2"], "kwarg": {"aaa": "bbb", "ccc": "ddd"}}
        ret = self.ckminions.spec_check(auth_list, "test.arg", args, "runner")
        self.assertTrue(ret)

        # Test list of args
        auth_list = [
            {
                "@runner": [
                    {
                        "test.arg": [
                            {
                                "args": ["1", "2"],
                                "kwargs": {"aaa": "bbb", "ccc": "ddd"},
                            },
                            {
                                "args": ["2", "3"],
                                "kwargs": {"aaa": "aaa", "ccc": "ccc"},
                            },
                        ]
                    }
                ]
            }
        ]
        args = {"arg": ["1", "2"], "kwarg": {"aaa": "bbb", "ccc": "ddd"}}
        ret = self.ckminions.spec_check(auth_list, "test.arg", args, "runner")
        self.assertTrue(ret)
        args = {"arg": ["2", "3"], "kwarg": {"aaa": "aaa", "ccc": "ccc"}}
        ret = self.ckminions.spec_check(auth_list, "test.arg", args, "runner")
        self.assertTrue(ret)

        # Test @module form
        auth_list = ["@jobs"]
        ret = self.ckminions.spec_check(auth_list, "jobs.active", {}, "runner")
        self.assertTrue(ret)
        ret = self.ckminions.spec_check(auth_list, "jobs.active", {}, "wheel")
        self.assertTrue(ret)
        ret = self.ckminions.spec_check(auth_list, "test.arg", {}, "runner")
        self.assertFalse(ret)
        ret = self.ckminions.spec_check(auth_list, "job.arg", {}, "runner")
        self.assertFalse(ret)

        # Test @module: function
        auth_list = [{"@jobs": "active"}]
        ret = self.ckminions.spec_check(auth_list, "jobs.active", {}, "runner")
        self.assertTrue(ret)
        ret = self.ckminions.spec_check(auth_list, "jobs.active", {}, "wheel")
        self.assertTrue(ret)
        ret = self.ckminions.spec_check(auth_list, "jobs.active_jobs", {}, "runner")
        self.assertTrue(ret)
        ret = self.ckminions.spec_check(auth_list, "jobs.activ", {}, "runner")
        self.assertFalse(ret)

        # Test @module: [functions]
        auth_list = [{"@jobs": ["active", "li"]}]
        ret = self.ckminions.spec_check(auth_list, "jobs.active", {}, "runner")
        self.assertTrue(ret)
        ret = self.ckminions.spec_check(auth_list, "jobs.list_jobs", {}, "runner")
        self.assertTrue(ret)
        ret = self.ckminions.spec_check(auth_list, "jobs.last_run", {}, "runner")
        self.assertFalse(ret)

        # Test @module: function with args
        auth_list = [
            {"@jobs": {"active": {"args": ["1", "2"], "kwargs": {"a": "b", "c": "d"}}}}
        ]
        args = {"arg": ["1", "2"], "kwarg": {"a": "b", "c": "d"}}
        ret = self.ckminions.spec_check(auth_list, "jobs.active", args, "runner")
        self.assertTrue(ret)
        ret = self.ckminions.spec_check(auth_list, "jobs.active", {}, "runner")
        self.assertFalse(ret)

    @patch(
        "salt.utils.minions.CkMinions._pki_minions",
        MagicMock(return_value=["alpha", "beta", "gamma"]),
    )
    def test_auth_check(self):
        # Test function-only rule
        auth_list = ["test.ping"]
        ret = self.ckminions.auth_check(auth_list, "test.ping", None, "alpha")
        self.assertTrue(ret)
        ret = self.ckminions.auth_check(auth_list, "test.arg", None, "alpha")
        self.assertFalse(ret)

        # Test minion and function
        auth_list = [{"alpha": "test.ping"}]
        ret = self.ckminions.auth_check(auth_list, "test.ping", None, "alpha")
        self.assertTrue(ret)
        ret = self.ckminions.auth_check(auth_list, "test.arg", None, "alpha")
        self.assertFalse(ret)
        ret = self.ckminions.auth_check(auth_list, "test.ping", None, "beta")
        self.assertFalse(ret)

        # Test function list
        auth_list = [{"*": ["test.*", "saltutil.cmd"]}]
        ret = self.ckminions.auth_check(auth_list, "test.arg", None, "alpha")
        self.assertTrue(ret)
        ret = self.ckminions.auth_check(auth_list, "test.ping", None, "beta")
        self.assertTrue(ret)
        ret = self.ckminions.auth_check(auth_list, "saltutil.cmd", None, "gamma")
        self.assertTrue(ret)
        ret = self.ckminions.auth_check(auth_list, "saltutil.running", None, "gamma")
        self.assertFalse(ret)

        # Test an args and kwargs rule
        auth_list = [
            {
                "alpha": {
                    "test.arg": {
                        "args": ["1", "2"],
                        "kwargs": {"aaa": "bbb", "ccc": "ddd"},
                    }
                }
            }
        ]
        ret = self.ckminions.auth_check(auth_list, "test.arg", None, "runner")
        self.assertFalse(ret)
        ret = self.ckminions.auth_check(auth_list, "test.arg", [], "runner")
        self.assertFalse(ret)
        args = ["1", "2", {"aaa": "bbb", "ccc": "ddd", "__kwarg__": True}]
        ret = self.ckminions.auth_check(auth_list, "test.arg", args, "runner")
        self.assertTrue(ret)
        args = [
            "1",
            "2",
            "3",
            {"aaa": "bbb", "ccc": "ddd", "eee": "fff", "__kwarg__": True},
        ]
        ret = self.ckminions.auth_check(auth_list, "test.arg", args, "runner")
        self.assertTrue(ret)
        args = ["1", {"aaa": "bbb", "ccc": "ddd", "__kwarg__": True}]
        ret = self.ckminions.auth_check(auth_list, "test.arg", args, "runner")
        self.assertFalse(ret)
        args = ["1", "2", {"aaa": "bbb", "__kwarg__": True}]
        ret = self.ckminions.auth_check(auth_list, "test.arg", args, "runner")
        self.assertFalse(ret)
        args = ["1", "3", {"aaa": "bbb", "ccc": "ddd", "__kwarg__": True}]
        ret = self.ckminions.auth_check(auth_list, "test.arg", args, "runner")
        self.assertFalse(ret)
        args = ["1", "2", {"aaa": "bbb", "ccc": "fff", "__kwarg__": True}]
        ret = self.ckminions.auth_check(auth_list, "test.arg", args, "runner")
        self.assertFalse(ret)

        # Test kwargs only rule
        auth_list = [{"alpha": {"test.arg": {"kwargs": {"aaa": "bbb", "ccc": "ddd"}}}}]
        args = ["1", "2", {"aaa": "bbb", "ccc": "ddd", "__kwarg__": True}]
        ret = self.ckminions.auth_check(auth_list, "test.arg", args, "runner")
        self.assertTrue(ret)
        args = [{"aaa": "bbb", "ccc": "ddd", "eee": "fff", "__kwarg__": True}]
        ret = self.ckminions.auth_check(auth_list, "test.arg", args, "runner")
        self.assertTrue(ret)

        # Test args only rule
        auth_list = [{"alpha": {"test.arg": {"args": ["1", "2"]}}}]
        args = ["1", "2", {"aaa": "bbb", "ccc": "ddd", "__kwarg__": True}]
        ret = self.ckminions.auth_check(auth_list, "test.arg", args, "runner")
        self.assertTrue(ret)
        args = ["1", "2"]
        ret = self.ckminions.auth_check(auth_list, "test.arg", args, "runner")
        self.assertTrue(ret)


class TargetParseTestCase(TestCase):
    def test_parse_grains_target(self):
        """
        Ensure proper parsing for grains
        """
        g_tgt = "G@a:b"
        ret = salt.utils.minions.parse_target(g_tgt)
        self.assertDictEqual(ret, {"engine": "G", "pattern": "a:b", "delimiter": None})

    def test_parse_grains_pcre_target(self):
        """
        Ensure proper parsing for grains PCRE matching
        """
        p_tgt = "P@a:b"
        ret = salt.utils.minions.parse_target(p_tgt)
        self.assertDictEqual(ret, {"engine": "P", "pattern": "a:b", "delimiter": None})

    def test_parse_pillar_pcre_target(self):
        """
        Ensure proper parsing for pillar PCRE matching
        """
        j_tgt = "J@a:b"
        ret = salt.utils.minions.parse_target(j_tgt)
        self.assertDictEqual(ret, {"engine": "J", "pattern": "a:b", "delimiter": None})

    def test_parse_list_target(self):
        """
        Ensure proper parsing for list matching
        """
        l_tgt = "L@a:b"
        ret = salt.utils.minions.parse_target(l_tgt)
        self.assertDictEqual(ret, {"engine": "L", "pattern": "a:b", "delimiter": None})

    def test_parse_nodegroup_target(self):
        """
        Ensure proper parsing for pillar matching
        """
        n_tgt = "N@a:b"
        ret = salt.utils.minions.parse_target(n_tgt)
        self.assertDictEqual(ret, {"engine": "N", "pattern": "a:b", "delimiter": None})

    def test_parse_subnet_target(self):
        """
        Ensure proper parsing for subnet matching
        """
        s_tgt = "S@a:b"
        ret = salt.utils.minions.parse_target(s_tgt)
        self.assertDictEqual(ret, {"engine": "S", "pattern": "a:b", "delimiter": None})

    def test_parse_minion_pcre_target(self):
        """
        Ensure proper parsing for minion PCRE matching
        """
        e_tgt = "E@a:b"
        ret = salt.utils.minions.parse_target(e_tgt)
        self.assertDictEqual(ret, {"engine": "E", "pattern": "a:b", "delimiter": None})

    def test_parse_range_target(self):
        """
        Ensure proper parsing for range matching
        """
        r_tgt = "R@a:b"
        ret = salt.utils.minions.parse_target(r_tgt)
        self.assertDictEqual(ret, {"engine": "R", "pattern": "a:b", "delimiter": None})

    def test_parse_multiword_target(self):
        """
        Ensure proper parsing for multi-word targets

        Refs https://github.com/saltstack/salt/issues/37231
        """
        mw_tgt = "G@a:b c"
        ret = salt.utils.minions.parse_target(mw_tgt)
        self.assertEqual(ret["pattern"], "a:b c")


class NodegroupCompTest(TestCase):
    """
    Test nodegroup comparisons found in
    salt.utils.minions.nodgroup_comp()
    """

    def test_simple_nodegroup(self):
        """
        Smoke test a very simple nodegroup. No recursion.
        """
        simple_nodegroup = {
            "group1": "L@foo.domain.com,bar.domain.com,baz.domain.com or bl*.domain.com"
        }

        ret = salt.utils.minions.nodegroup_comp("group1", simple_nodegroup)
        expected_ret = [
            "L@foo.domain.com,bar.domain.com,baz.domain.com",
            "or",
            "bl*.domain.com",
        ]
        self.assertListEqual(ret, expected_ret)

    def test_simple_expression_nodegroup(self):
        """
        Smoke test a nodegroup with a simple expression. No recursion.
        """
        simple_nodegroup = {"group1": "[foo,bar,baz].domain.com"}

        ret = salt.utils.minions.nodegroup_comp("group1", simple_nodegroup)
        expected_ret = ["E@[foo,bar,baz].domain.com"]
        self.assertListEqual(ret, expected_ret)

    def test_simple_recurse(self):
        """
        Test a case where one nodegroup contains a second nodegroup
        """
        referenced_nodegroups = {
            "group1": (
                "L@foo.domain.com,bar.domain.com,baz.domain.com or bl*.domain.com"
            ),
            "group2": "G@os:Debian and N@group1",
        }

        ret = salt.utils.minions.nodegroup_comp("group2", referenced_nodegroups)
        expected_ret = [
            "(",
            "G@os:Debian",
            "and",
            "(",
            "L@foo.domain.com,bar.domain.com,baz.domain.com",
            "or",
            "bl*.domain.com",
            ")",
            ")",
        ]
        self.assertListEqual(ret, expected_ret)

    def test_circular_nodegroup_reference(self):
        """
        Test to see what happens if A refers to B
        and B in turn refers back to A
        """
        referenced_nodegroups = {"group1": "N@group2", "group2": "N@group1"}

        # If this works, it should also print an error to the console
        ret = salt.utils.minions.nodegroup_comp("group1", referenced_nodegroups)
        self.assertEqual(ret, [])



import pytest

pytestmark = [
    pytest.mark.skip_on_windows,
]


def test_grains_items(salt_cli, salt_minion, salt_master):
    """
    Test grains.items
    """
    assert salt_master.is_running()

    ret = salt_cli.run("grains.items", minion_tgt=salt_minion.id)
    assert ret.data, ret
    assert "osrelease" in ret.data


def test_grains_item_os(salt_cli, salt_minion, salt_master):
    """
    Test grains.item os
    """
    assert salt_master.is_running()

    ret = salt_cli.run("grains.item", "os", minion_tgt=salt_minion.id)
    assert ret.data, ret
    assert "os" in ret.data


def test_grains_item_pythonversion(salt_cli, salt_minion, salt_master):
    """
    Test grains.item pythonversion
    """
    assert salt_master.is_running()

    ret = salt_cli.run("grains.item", "pythonversion", minion_tgt=salt_minion.id)
    assert ret.data, ret
    assert "pythonversion" in ret.data


def test_grains_setval_key_val(salt_cli, salt_minion, salt_master):
    """
    Test grains.setval key val
    """
    assert salt_master.is_running()

    ret = salt_cli.run("grains.setval", "key", "val", minion_tgt=salt_minion.id)
    assert ret.data, ret
    assert "key" in ret.data



"""
Functions to interact with the pillar compiler on the master
"""

import logging

import salt.pillar
import salt.utils.minions

log = logging.getLogger(__name__)


def show_top(minion=None, saltenv="base"):
    """
    Returns the compiled top data for pillar for a specific minion.  If no
    minion is specified, we use the first minion we find.

    CLI Example:

    .. code-block:: bash

        salt-run pillar.show_top
    """
    id_, grains, _ = salt.utils.minions.get_minion_data(minion, __opts__)
    pillar = salt.pillar.Pillar(__opts__, grains, id_, saltenv)

    top, errors = pillar.get_top()

    if errors:
        __jid_event__.fire_event({"data": errors, "outputter": "nested"}, "progress")
        return errors

    return top


def show_pillar(minion="*", **kwargs):
    """
    Returns the compiled pillar either of a specific minion
    or just the global available pillars. This function assumes
    that no minion has the id ``*``.
    Function also accepts pillarenv as attribute in order to limit to a specific pillar branch of git

    CLI Example:

    shows minion specific pillar:

    .. code-block:: bash

        salt-run pillar.show_pillar 'www.example.com'

    shows global pillar:

    .. code-block:: bash

        salt-run pillar.show_pillar

    shows global pillar for 'dev' pillar environment:
    (note that not specifying pillarenv will merge all pillar environments
    using the master config option pillar_source_merging_strategy.)

    .. code-block:: bash

        salt-run pillar.show_pillar 'pillarenv=dev'

    shows global pillar for 'dev' pillar environment and specific pillarenv = dev:

    .. code-block:: bash

        salt-run pillar.show_pillar 'saltenv=dev' 'pillarenv=dev'

    API Example:

    .. code-block:: python

        import salt.config
        import salt.runner
        opts = salt.config.master_config('/etc/salt/master')
        runner = salt.runner.RunnerClient(opts)
        pillar = runner.cmd('pillar.show_pillar', [])
        print(pillar)
    """
    pillarenv = None
    saltenv = "base"
    id_, grains, _ = salt.utils.minions.get_minion_data(minion, __opts__)
    if grains is None:
        grains = {"fqdn": minion}

    for key in kwargs:
        if key == "saltenv":
            saltenv = kwargs[key]
        elif key == "pillarenv":
            # pillarenv overridden on CLI
            pillarenv = kwargs[key]
        else:
            grains[key] = kwargs[key]

    pillar = salt.pillar.Pillar(__opts__, grains, id_, saltenv, pillarenv=pillarenv)

    compiled_pillar = pillar.compile_pillar()
    return compiled_pillar


def clear_pillar_cache(minion="*", **kwargs):
    """
    Clears the cached values when using pillar_cache

    .. versionadded:: 3003

    CLI Example:

    Clears the pillar cache for a specific minion:

    .. code-block:: bash

        salt-run pillar.clear_pillar_cache 'minion'

    """

    if not __opts__.get("pillar_cache"):
        log.info("The pillar_cache is set to False or not enabled.")
        return False

    ckminions = salt.utils.minions.CkMinions(__opts__)
    ret = ckminions.check_minions(minion)

    pillarenv = kwargs.pop("pillarenv", None)
    saltenv = kwargs.pop("saltenv", "base")

    pillar_cache = {}
    for tgt in ret.get("minions", []):
        id_, grains, _ = salt.utils.minions.get_minion_data(tgt, __opts__)

        for key in kwargs:
            grains[key] = kwargs[key]

        if grains is None:
            grains = {"fqdn": minion}

        pillar = salt.pillar.PillarCache(
            __opts__, grains, id_, saltenv, pillarenv=pillarenv
        )
        pillar.clear_pillar()

        if __opts__.get("pillar_cache_backend") == "memory":
            _pillar_cache = pillar.cache
        else:
            _pillar_cache = pillar.cache._dict

        if tgt in _pillar_cache and _pillar_cache[tgt]:
            pillar_cache[tgt] = _pillar_cache.get(tgt).get(pillarenv)

    return pillar_cache


def show_pillar_cache(minion="*", **kwargs):
    """
    Shows the cached values in pillar_cache

    .. versionadded:: 3003

    CLI Example:

    Shows the pillar cache for a specific minion:

    .. code-block:: bash

        salt-run pillar.show_pillar_cache 'minion'

    """

    if not __opts__.get("pillar_cache"):
        log.info("The pillar_cache is set to False or not enabled.")
        return False

    ckminions = salt.utils.minions.CkMinions(__opts__)
    ret = ckminions.check_minions(minion)

    pillarenv = kwargs.pop("pillarenv", None)
    saltenv = kwargs.pop("saltenv", "base")

    pillar_cache = {}
    for tgt in ret.get("minions", []):
        id_, grains, _ = salt.utils.minions.get_minion_data(tgt, __opts__)

        for key in kwargs:
            grains[key] = kwargs[key]

        if grains is None:
            grains = {"fqdn": minion}

        pillar = salt.pillar.PillarCache(
            __opts__, grains, id_, saltenv, pillarenv=pillarenv
        )

        if __opts__.get("pillar_cache_backend") == "memory":
            _pillar_cache = pillar.cache
        else:
            _pillar_cache = pillar.cache._dict

        if tgt in _pillar_cache and _pillar_cache[tgt]:
            pillar_cache[tgt] = _pillar_cache[tgt].get(pillarenv)

    return pillar_cache
